from metaflow._vendor import click

from .. import parameters
from ..runtime import NativeRuntime


@parameters.add_custom_parameters(deploy_mode=False)
@click.command(help="Internal command to initialize a run.", hidden=True)
@click.option(
    "--run-id",
    default=None,
    required=True,
    help="ID for one execution of all steps in the flow.",
)
@click.option(
    "--task-id", default=None, required=True, help="ID for this instance of the step."
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    default=None,
    help="Tags for this instance of the step.",
)
@click.pass_obj
def init(obj, run_id=None, task_id=None, tags=None, **kwargs):
    # init is a separate command instead of an option in 'step'
    # since we need to capture user-specified parameters with
    # @add_custom_parameters. Adding custom parameters to 'step'
    # is not desirable due to the possibility of name clashes between
    # user-specified parameters and our internal options. Note that
    # user-specified parameters are often defined as environment
    # variables.

    obj.metadata.add_sticky_tags(tags=tags)

    runtime = NativeRuntime(
        obj.flow,
        obj.graph,
        obj.flow_datastore,
        obj.metadata,
        obj.environment,
        obj.package,
        obj.logger,
        obj.entrypoint,
        obj.event_logger,
        obj.monitor,
        run_id=run_id,
        skip_decorator_hooks=True,
    )
    obj.flow._set_constants(obj.graph, kwargs, obj.config_options)
    runtime.persist_constants(task_id=task_id)



from __future__ import print_function

import errno
import json
import time
import math
import random
import re
import sys
import os
import traceback
from collections import namedtuple
from functools import partial, wraps
from hashlib import sha1
from tempfile import NamedTemporaryFile
from multiprocessing import Process, Queue
from itertools import starmap, chain, islice

from boto3.exceptions import RetriesExceededError, S3UploadFailedError
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError, SSLError

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

if __name__ == "__main__":
    # When launched standalone, point to our parent metaflow
    sys.path.insert(
        0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../../"))
    )

from metaflow._vendor import click

# we use Metaflow's parallel_imap_unordered instead of
# multiprocessing.Pool because https://bugs.python.org/issue31886
from metaflow.util import TempDir, url_quote, url_unquote
from metaflow.multicore_utils import parallel_map
from metaflow.plugins.datatools.s3.s3util import (
    aws_retry,
    read_in_chunks,
    get_timestamp,
    TRANSIENT_RETRY_LINE_CONTENT,
    TRANSIENT_RETRY_START_LINE,
)
import metaflow.tracing as tracing
from metaflow.metaflow_config import (
    S3_WORKER_COUNT,
)

DOWNLOAD_FILE_THRESHOLD = 2 * TransferConfig().multipart_threshold
DOWNLOAD_MAX_CHUNK = 2 * 1024 * 1024 * 1024 - 1

RANGE_MATCH = re.compile(r"bytes (?P<start>[0-9]+)-(?P<end>[0-9]+)/(?P<total>[0-9]+)")

# from botocore ClientError MSG_TEMPLATE:
# https://github.com/boto/botocore/blob/68ca78f3097906c9231840a49931ef4382c41eea/botocore/exceptions.py#L521
BOTOCORE_MSG_TEMPLATE_MATCH = re.compile(
    r"An error occurred \((\w+)\) when calling the (\w+) operation.*: (.+)"
)

S3Config = namedtuple("S3Config", "role session_vars client_params")


class S3Url(object):
    def __init__(
        self,
        bucket,
        path,
        url,
        local,
        prefix,
        content_type=None,
        encryption=None,
        metadata=None,
        range=None,
        idx=None,
    ):

        self.bucket = bucket
        self.path = path
        self.url = url
        self.local = local
        self.prefix = prefix
        self.content_type = content_type
        self.metadata = metadata
        self.range = range
        self.idx = idx
        self.encryption = encryption

    def __str__(self):
        return self.url


# We use error codes instead of Exceptions, which are trickier to
# handle reliably in a multiprocess world
ERROR_INVALID_URL = 4
ERROR_NOT_FULL_PATH = 5
ERROR_URL_NOT_FOUND = 6
ERROR_URL_ACCESS_DENIED = 7
ERROR_WORKER_EXCEPTION = 8
ERROR_VERIFY_FAILED = 9
ERROR_LOCAL_FILE_NOT_FOUND = 10
ERROR_INVALID_RANGE = 11
ERROR_TRANSIENT = 12
ERROR_OUT_OF_DISK_SPACE = 13


def format_result_line(idx, prefix, url="", local=""):
    # We prefix each output with the index corresponding to the line number on the
    # initial request (ie: prior to any transient errors). This allows us to
    # properly maintain the order in which things were requested even in the presence
    # of transient retries where we do not know what succeeds and what does not.
    # Basically, when we retry an operation, we can trace it back to its original
    # position in the first request.
    return " ".join(
        [str(idx)] + [url_quote(x).decode("utf-8") for x in (prefix, url, local)]
    )


# I can't understand what's the right way to deal
# with boto errors. This function can be replaced
# with better error handling code.
def normalize_client_error(err):
    error_code = err.response["Error"]["Code"]
    try:
        return int(error_code)
    except ValueError:
        if error_code in ("AccessDenied", "AllAccessDisabled", "InvalidAccessKeyId"):
            return 403
        if error_code in ("NoSuchKey", "NoSuchBucket"):
            return 404
        if error_code == "InvalidRange":
            return 416
        # We "normalize" retriable server errors to 503. These are also considered
        # transient by boto3 (see:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html)
        if error_code in (
            "SlowDown",
            "RequestTimeout",
            "RequestTimeoutException",
            "PriorRequestNotComplete",
            "ConnectionError",
            "HTTPClientError",
            "Throttling",
            "ThrottlingException",
            "ThrottledException",
            "RequestThrottledException",
            "TooManyRequestsException",
            "ProvisionedThroughputExceededException",
            "TransactionInProgressException",
            "RequestLimitExceeded",
            "BandwidthLimitExceeded",
            "LimitExceededException",
            "RequestThrottled",
            "EC2ThrottledException",
            "InternalError",
        ):
            return 503
    return error_code


# S3 worker pool


@tracing.cli("s3op/worker")
def worker(result_file_name, queue, mode, s3config):
    # Interpret mode, it can either be a single op or something like
    # info_download or info_upload which implies:
    #  - for download: we need to return the information as well
    #  - for upload: we need to not overwrite the file if it exists
    modes = mode.split("_")
    pre_op_info = False
    if len(modes) > 1:
        pre_op_info = True
        mode = modes[1]
    else:
        mode = modes[0]

    def op_info(url):
        try:
            head = s3.head_object(Bucket=url.bucket, Key=url.path)
            to_return = {
                "error": None,
                "size": head["ContentLength"],
                "content_type": head["ContentType"],
                "encryption": head.get("ServerSideEncryption"),
                "metadata": head["Metadata"],
                "last_modified": get_timestamp(head["LastModified"]),
            }
        except client_error as err:
            error_code = normalize_client_error(err)
            if error_code == 404:
                to_return = {"error": ERROR_URL_NOT_FOUND, "raise_error": err}
            elif error_code == 403:
                to_return = {"error": ERROR_URL_ACCESS_DENIED, "raise_error": err}
            elif error_code == 416:
                to_return = {"error": ERROR_INVALID_RANGE, "raise_error": err}
            elif error_code in (500, 502, 503, 504):
                to_return = {"error": ERROR_TRANSIENT, "raise_error": err}
            else:
                to_return = {"error": error_code, "raise_error": err}
        return to_return

    with open(result_file_name, "w") as result_file:
        try:
            from metaflow.plugins.datatools.s3.s3util import get_s3_client

            s3, client_error = get_s3_client(
                s3_role_arn=s3config.role,
                s3_session_vars=s3config.session_vars,
                s3_client_params=s3config.client_params,
            )
            while True:
                url, idx = queue.get()
                if url is None:
                    break
                if mode == "info":
                    result = op_info(url)
                    orig_error = result.get("raise_error", None)
                    if orig_error:
                        del result["raise_error"]
                    with open(url.local, "w") as f:
                        json.dump(result, f)
                    result_file.write(
                        "%d %d\n"
                        % (idx, -1 * result["error"] if orig_error else result["size"])
                    )
                elif mode == "download":
                    tmp = NamedTemporaryFile(dir=".", mode="wb", delete=False)
                    try:
                        try:
                            if url.range:
                                resp = s3.get_object(
                                    Bucket=url.bucket, Key=url.path, Range=url.range
                                )
                                range_result = resp["ContentRange"]
                                range_result_match = RANGE_MATCH.match(range_result)
                                if range_result_match is None:
                                    raise RuntimeError(
                                        "Wrong format for ContentRange: %s"
                                        % str(range_result)
                                    )
                                range_result = {
                                    x: int(range_result_match.group(x))
                                    for x in ["total", "start", "end"]
                                }
                            else:
                                resp = s3.get_object(Bucket=url.bucket, Key=url.path)
                                range_result = None
                            sz = resp["ContentLength"]
                            if range_result is None:
                                range_result = {"total": sz, "start": 0, "end": sz - 1}
                            if not url.range and sz > DOWNLOAD_FILE_THRESHOLD:
                                # In this case, it is more efficient to use download_file as it
                                # will download multiple parts in parallel (it does it after
                                # multipart_threshold)
                                s3.download_file(url.bucket, url.path, tmp.name)
                            else:
                                read_in_chunks(
                                    tmp, resp["Body"], sz, DOWNLOAD_MAX_CHUNK
                                )
                            tmp.close()
                            os.rename(tmp.name, url.local)
                        except client_error as err:
                            tmp.close()
                            os.unlink(tmp.name)
                            handle_client_error(err, idx, result_file)
                            continue
                        except RetriesExceededError as e:
                            tmp.close()
                            os.unlink(tmp.name)
                            err = convert_to_client_error(e)
                            handle_client_error(err, idx, result_file)
                            continue
                        except OSError as e:
                            tmp.close()
                            os.unlink(tmp.name)
                            if e.errno == errno.ENOSPC:
                                result_file.write(
                                    "%d %d\n" % (idx, -ERROR_OUT_OF_DISK_SPACE)
                                )
                            else:
                                result_file.write("%d %d\n" % (idx, -ERROR_TRANSIENT))
                            result_file.flush()
                            continue
                    except (SSLError, Exception) as e:
                        tmp.close()
                        os.unlink(tmp.name)
                        # assume anything else is transient
                        result_file.write("%d %d\n" % (idx, -ERROR_TRANSIENT))
                        result_file.flush()
                        continue
                    # If we need the metadata, get it and write it out
                    if pre_op_info:
                        with open("%s_meta" % url.local, mode="w") as f:
                            # Get range information

                            args = {
                                "size": resp["ContentLength"],
                                "range_result": range_result,
                            }
                            if resp["ContentType"]:
                                args["content_type"] = resp["ContentType"]
                            if resp["Metadata"] is not None:
                                args["metadata"] = resp["Metadata"]
                            if resp.get("ServerSideEncryption") is not None:
                                args["encryption"] = resp["ServerSideEncryption"]
                            if resp["LastModified"]:
                                args["last_modified"] = get_timestamp(
                                    resp["LastModified"]
                                )
                            json.dump(args, f)
                    # Finally, we push out the size to the result_pipe since
                    # the size is used for verification and other purposes, and
                    # we want to avoid file operations for this simple process
                    result_file.write("%d %d\n" % (idx, resp["ContentLength"]))
                else:
                    # This is upload, if we have a pre_op, it means we do not
                    # want to overwrite
                    do_upload = False
                    if pre_op_info:
                        result_info = op_info(url)
                        if result_info["error"] == ERROR_URL_NOT_FOUND:
                            # We only upload if the file is not found
                            do_upload = True
                    else:
                        # No pre-op so we upload
                        do_upload = True
                    if do_upload:
                        extra = None
                        if url.content_type or url.metadata or url.encryption:
                            extra = {}
                            if url.content_type:
                                extra["ContentType"] = url.content_type
                            if url.metadata is not None:
                                extra["Metadata"] = url.metadata
                            if url.encryption is not None:
                                extra["ServerSideEncryption"] = url.encryption
                        try:
                            try:
                                s3.upload_file(
                                    url.local, url.bucket, url.path, ExtraArgs=extra
                                )
                                # We indicate that the file was uploaded
                                result_file.write("%d %d\n" % (idx, 0))
                            except client_error as err:
                                # Shouldn't get here, but just in case.
                                # Internally, botocore catches ClientError and returns a S3UploadFailedError.
                                # See https://github.com/boto/boto3/blob/develop/boto3/s3/transfer.py#L377
                                handle_client_error(err, idx, result_file)
                                continue
                            except S3UploadFailedError as e:
                                err = convert_to_client_error(e)
                                handle_client_error(err, idx, result_file)
                                continue
                        except (SSLError, Exception) as e:
                            # assume anything else is transient
                            result_file.write("%d %d\n" % (idx, -ERROR_TRANSIENT))
                            result_file.flush()
                            continue
        except:
            traceback.print_exc()
            result_file.flush()
            sys.exit(ERROR_WORKER_EXCEPTION)


def convert_to_client_error(e):
    match = BOTOCORE_MSG_TEMPLATE_MATCH.search(str(e))
    if not match:
        raise e
    error_code = match.group(1)
    operation_name = match.group(2)
    error_message = match.group(3)
    response = {
        "Error": {
            "Code": error_code,
            "Message": error_message,
        }
    }
    return ClientError(response, operation_name)


def handle_client_error(err, idx, result_file):
    error_code = normalize_client_error(err)
    if error_code == 404:
        result_file.write("%d %d\n" % (idx, -ERROR_URL_NOT_FOUND))
        result_file.flush()
    elif error_code == 403:
        result_file.write("%d %d\n" % (idx, -ERROR_URL_ACCESS_DENIED))
        result_file.flush()
    elif error_code == 503:
        result_file.write("%d %d\n" % (idx, -ERROR_TRANSIENT))
        result_file.flush()
    else:
        # optimistically assume it is a transient error
        result_file.write("%d %d\n" % (idx, -ERROR_TRANSIENT))
        result_file.flush()
    # TODO specific error message for out of disk space


def start_workers(mode, urls, num_workers, inject_failure, s3config):
    # We start the minimum of len(urls) or num_workers to avoid starting
    # workers that will definitely do nothing
    num_workers = min(num_workers, len(urls))
    queue = Queue(len(urls) + num_workers)
    procs = {}
    random.seed()

    sz_results = []
    # 1. push sources and destinations to the queue
    # We only push if we don't inject a failure; otherwise, we already set the sz_results
    # appropriately with the result of the injected failure.
    for idx, elt in enumerate(urls):
        if random.randint(0, 99) < inject_failure:
            sz_results.append(-ERROR_TRANSIENT)
        else:
            sz_results.append(None)
            queue.put((elt, idx))

    # 2. push end-of-queue markers
    for i in range(num_workers):
        queue.put((None, None))

    # 3. start processes
    with TempDir() as output_dir:
        for i in range(num_workers):
            file_path = os.path.join(output_dir, str(i))
            p = Process(
                target=worker,
                args=(file_path, queue, mode, s3config),
            )
            p.start()
            procs[p] = file_path

        # 4. wait for the processes to finish; we continuously update procs
        # to remove all processes that have finished already
        while procs:
            new_procs = {}
            for proc, out_path in procs.items():
                proc.join(timeout=1)
                if proc.exitcode is not None:
                    if proc.exitcode != 0:
                        msg = "Worker process failed (exit code %d)" % proc.exitcode

                        # IMPORTANT: if this process has put items on a queue, then it will not terminate
                        # until all buffered items have been flushed to the pipe, causing a deadlock.
                        # `cancel_join_thread()` allows it to exit without flushing the queue.
                        # Without this line, the parent process would hang indefinitely when a subprocess
                        # did not exit cleanly in the case of unhandled exceptions.
                        #
                        # The error situation is:
                        # 1. this process puts stuff in queue
                        # 2. subprocess dies so doesn't consume its end-of-queue marker (the None)
                        # 3. other subprocesses consume all useful bits AND their end-of-queue marker
                        # 4. one marker is left and not consumed
                        # 5. this process cannot shut down until the queue is empty.
                        # 6. it will never be empty because all subprocesses (workers) have died.
                        queue.cancel_join_thread()

                        exit(msg, proc.exitcode)
                    # Read the output file if all went well
                    with open(out_path, "r") as out_file:
                        for line in out_file:
                            line_split = line.split(" ")
                            sz_results[int(line_split[0])] = int(line_split[1])
                else:
                    # Put this process back in the processes to check
                    new_procs[proc] = out_path
            procs = new_procs
    return sz_results


def process_urls(mode, urls, verbose, inject_failure, num_workers, s3config):

    if verbose:
        print("%sing %d files.." % (mode.capitalize(), len(urls)), file=sys.stderr)

    start = time.time()
    sz_results = start_workers(mode, urls, num_workers, inject_failure, s3config)
    end = time.time()

    if verbose:
        total_size = sum(sz for sz in sz_results if sz is not None and sz > 0)
        bw = total_size / (end - start)
        print(
            "%sed %d files, %s in total, in %d seconds (%s/s)."
            % (
                mode.capitalize(),
                len(urls),
                with_unit(total_size),
                end - start,
                with_unit(bw),
            ),
            file=sys.stderr,
        )
    return sz_results


# Utility functions


def with_unit(x):
    if x > 1024**3:
        return "%.1fGB" % (x / 1024.0**3)
    elif x > 1024**2:
        return "%.1fMB" % (x / 1024.0**2)
    elif x > 1024:
        return "%.1fKB" % (x / 1024.0)
    else:
        return "%d bytes" % x


# S3Ops class is just a wrapper for get_size and list_prefix
# required by @aws_retry decorator, which needs the reset_client
# method. Otherwise they would be just stand-alone functions.
class S3Ops(object):
    def __init__(self, s3config):
        self.s3 = None
        self.s3config = s3config
        self.client_error = None

    def reset_client(self, hard_reset=False):
        from metaflow.plugins.datatools.s3.s3util import get_s3_client

        if hard_reset or self.s3 is None:
            self.s3, self.client_error = get_s3_client(
                s3_role_arn=self.s3config.role,
                s3_session_vars=self.s3config.session_vars,
                s3_client_params=self.s3config.client_params,
            )

    @aws_retry
    def get_info(self, url):
        self.reset_client()
        try:
            head = self.s3.head_object(Bucket=url.bucket, Key=url.path)
            return (
                True,
                url,
                [
                    (
                        S3Url(
                            bucket=url.bucket,
                            path=url.path,
                            url=url.url,
                            local=url.local,
                            prefix=url.prefix,
                            content_type=head["ContentType"],
                            metadata=head["Metadata"],
                            encryption=head.get("ServerSideEncryption"),
                            range=url.range,
                        ),
                        head["ContentLength"],
                    )
                ],
            )
        except self.client_error as err:
            error_code = normalize_client_error(err)
            if error_code == 404:
                return False, url, ERROR_URL_NOT_FOUND
            elif error_code == 403:
                return False, url, ERROR_URL_ACCESS_DENIED
            # Transient errors are going to be retried by the aws_retry decorator
            else:
                raise

    @aws_retry
    def list_prefix(self, prefix_url, delimiter=""):
        self.reset_client()
        url_base = "s3://%s/" % prefix_url.bucket
        try:
            paginator = self.s3.get_paginator("list_objects_v2")
            urls = []
            for page in paginator.paginate(
                Bucket=prefix_url.bucket, Prefix=prefix_url.path, Delimiter=delimiter
            ):
                # note that an url may be both a prefix and an object
                # - the trailing slash is significant in S3
                if "Contents" in page:
                    for key in page.get("Contents", []):
                        url = url_base + key["Key"]
                        urlobj = S3Url(
                            url=url,
                            bucket=prefix_url.bucket,
                            path=key["Key"],
                            local=generate_local_path(url),
                            prefix=prefix_url.url,
                        )
                        urls.append((urlobj, key["Size"]))
                if "CommonPrefixes" in page:
                    # we get CommonPrefixes if Delimiter is a non-empty string
                    for key in page.get("CommonPrefixes", []):
                        url = url_base + key["Prefix"]
                        urlobj = S3Url(
                            url=url,
                            bucket=prefix_url.bucket,
                            path=key["Prefix"],
                            local=None,
                            prefix=prefix_url.url,
                        )
                        urls.append((urlobj, None))
            return True, prefix_url, urls
        except self.s3.exceptions.NoSuchBucket:
            return False, prefix_url, ERROR_URL_NOT_FOUND
        except self.client_error as err:
            error_code = normalize_client_error(err)
            if error_code == 404:
                return False, prefix_url, ERROR_URL_NOT_FOUND
            elif error_code == 403:
                return False, prefix_url, ERROR_URL_ACCESS_DENIED
            # Transient errors are going to be retried by the aws_retry decorator
            else:
                raise


# We want to reuse an S3 client instance over multiple operations.
# This is accomplished by op_ functions below.


def op_get_info(s3config, urls):
    s3 = S3Ops(s3config)
    return [s3.get_info(url) for url in urls]


def op_list_prefix(s3config, prefix_urls):
    s3 = S3Ops(s3config)
    return [s3.list_prefix(prefix) for prefix in prefix_urls]


def op_list_prefix_nonrecursive(s3config, prefix_urls):
    s3 = S3Ops(s3config)
    return [s3.list_prefix(prefix, delimiter="/") for prefix in prefix_urls]


def exit(exit_code, url):
    if exit_code == ERROR_INVALID_URL:
        msg = "Invalid url: %s" % url.url
    elif exit_code == ERROR_NOT_FULL_PATH:
        msg = "URL not a full path: %s" % url.url
    elif exit_code == ERROR_URL_NOT_FOUND:
        msg = "URL not found: %s" % url.url
    elif exit_code == ERROR_URL_ACCESS_DENIED:
        msg = "Access denied to URL: %s" % url.url
    elif exit_code == ERROR_WORKER_EXCEPTION:
        msg = "Download failed"
    elif exit_code == ERROR_VERIFY_FAILED:
        msg = "Verification failed for URL %s, local file %s" % (url.url, url.local)
    elif exit_code == ERROR_LOCAL_FILE_NOT_FOUND:
        msg = "Local file not found: %s" % url
    elif exit_code == ERROR_TRANSIENT:
        msg = "Transient error for url: %s" % url
    elif exit_code == ERROR_OUT_OF_DISK_SPACE:
        msg = "Out of disk space when downloading URL: %s" % url
    else:
        msg = "Unknown error"
    print("s3op failed:\n%s" % msg, file=sys.stderr)
    sys.exit(exit_code)


def verify_results(urls, verbose=False):
    for url, expected in urls:
        if verbose:
            print("verifying %s, expected %s" % (url, expected), file=sys.stderr)
        try:
            got = os.stat(url.local).st_size
        except OSError:
            raise
        if expected != got:
            exit(ERROR_VERIFY_FAILED, url)
        if url.content_type or url.metadata or url.encryption:
            # Verify that we also have a metadata file present
            try:
                os.stat("%s_meta" % url.local)
            except OSError:
                exit(ERROR_VERIFY_FAILED, url)


def generate_local_path(url, range="whole", suffix=None):
    # this function generates a safe local file name corresponding to
    # an S3 URL. URLs may be longer than maximum file length limit on Linux,
    # so we mostly hash the URL but retain the leaf part as a convenience
    # feature to ease eyeballing
    # We also call out "range" specifically to allow multiple ranges for the same
    # file to be downloaded in parallel.
    if range is None:
        range = "whole"
    if range != "whole":
        # It will be of the form `bytes=%d-` or `bytes=-%d` or `bytes=%d-%d`
        range = range[6:].replace("-", "_")
    quoted = url_quote(url)
    fname = quoted.split(b"/")[-1].replace(b".", b"_").replace(b"-", b"_")
    sha = sha1(quoted).hexdigest()
    if suffix:
        return "-".join((sha, fname.decode("utf-8"), range, suffix))
    return "-".join((sha, fname.decode("utf-8"), range))


def parallel_op(op, lst, num_workers):
    # parallel op divides work equally amongst num_workers
    # processes. This is a good strategy if the cost is
    # uniform over the units of work, e.g. op_get_info, which
    # is a single HEAD request to S3.
    #
    # This approach is less optimal with op_list_prefix where
    # the cost of S3 listing per prefix can vary drastically.
    # We could optimize this case by using a worker model with
    # a queue, like for downloads but the difference here is
    # that we need to return a value, which would require a
    # bit more work - something to consider if this turns out
    # to be a bottleneck.
    if lst:
        num = min(len(lst), num_workers)
        batch_size = math.ceil(len(lst) / float(num))
        batches = []
        it = iter(lst)
        while True:
            batch = list(islice(it, batch_size))
            if batch:
                batches.append(batch)
            else:
                break
        it = parallel_map(op, batches, max_parallel=num)
        for x in chain.from_iterable(it):
            yield x


# CLI


def common_options(func):
    @click.option(
        "--inputs",
        type=click.Path(exists=True),
        help="Read input prefixes from the given file.",
    )
    @click.option(
        "--num-workers",
        default=S3_WORKER_COUNT,
        show_default=True,
        help="Number of concurrent connections.",
    )
    @click.option(
        "--s3role",
        default=None,
        show_default=True,
        required=False,
        help="Role to assume when getting the S3 client",
    )
    @click.option(
        "--s3sessionvars",
        default=None,
        show_default=True,
        required=False,
        help="Session vars to set when getting the S3 client",
    )
    @click.option(
        "--s3clientparams",
        default=None,
        show_default=True,
        required=False,
        help="Client parameters to set when getting the S3 client",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def non_lst_common_options(func):
    @click.option(
        "--verbose/--no-verbose",
        default=True,
        show_default=True,
        help="Print status information on stderr.",
    )
    @click.option(
        "--listing/--no-listing",
        default=False,
        show_default=True,
        help="Print S3 URL -> local file mapping on stdout.",
    )
    @click.option(
        "--inject-failure",
        default=0,
        show_default=True,
        type=int,
        help="Simulate transient failures -- percentage (int) of injected failures",
        hidden=True,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group()
def cli():
    pass


@cli.command("list", help="List S3 objects")
@tracing.cli("s3op/list")
@click.option(
    "--recursive/--no-recursive",
    default=False,
    show_default=True,
    help="List prefixes recursively.",
)
@common_options
@click.argument("prefixes", nargs=-1)
def lst(
    prefixes,
    inputs=None,
    num_workers=None,
    recursive=None,
    s3role=None,
    s3sessionvars=None,
    s3clientparams=None,
):

    s3config = S3Config(
        s3role,
        json.loads(s3sessionvars) if s3sessionvars else None,
        json.loads(s3clientparams) if s3clientparams else None,
    )

    urllist = []
    to_iterate, _ = _populate_prefixes(prefixes, inputs)
    for _, prefix, url, _ in to_iterate:
        src = urlparse(url)
        url = S3Url(
            url=url,
            bucket=src.netloc,
            path=src.path.lstrip("/"),
            local=None,
            prefix=prefix,
        )
        if src.scheme != "s3":
            exit(ERROR_INVALID_URL, url)
        urllist.append(url)

    op = (
        partial(op_list_prefix, s3config)
        if recursive
        else partial(op_list_prefix_nonrecursive, s3config)
    )
    urls = []
    for success, prefix_url, ret in parallel_op(op, urllist, num_workers):
        if success:
            urls.extend(ret)
        else:
            exit(ret, prefix_url)

    for idx, (url, size) in enumerate(urls):
        if size is None:
            print(format_result_line(idx, url.prefix, url.url))
        else:
            print(format_result_line(idx, url.prefix, url.url, str(size)))


@cli.command(help="Upload files to S3")
@tracing.cli("s3op/put")
@click.option(
    "--file",
    "files",
    type=(click.Path(exists=True), str),
    multiple=True,
    help="Local file->S3Url pair to upload. Can be specified multiple times.",
)
@click.option(
    "--filelist",
    type=click.Path(exists=True),
    help="Read local file -> S3 URL mappings from the given file. Use --inputs instead",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=True,
    show_default=True,
    help="Overwrite key if it already exists in S3.",
)
@common_options
@non_lst_common_options
def put(
    files=None,
    filelist=None,
    inputs=None,
    num_workers=None,
    verbose=None,
    overwrite=True,
    listing=None,
    s3role=None,
    s3sessionvars=None,
    s3clientparams=None,
    inject_failure=0,
):
    if inputs is not None and filelist is not None:
        raise RuntimeError("Cannot specify inputs and filelist at the same time")
    if inputs is not None and filelist is None:
        filelist = inputs

    is_transient_retry = False

    def _files():
        nonlocal is_transient_retry
        line_idx = 0
        for local, url in files:
            local_file = url_unquote(local)
            if not os.path.exists(local_file):
                exit(ERROR_LOCAL_FILE_NOT_FOUND, local_file)
            yield line_idx, local_file, url_unquote(url), None, None
            line_idx += 1
        if filelist:
            # NOTE: We are assuming that the idx is properly set. This is only used
            # by the transient failure retry mechanism and users should not use it
            # directly. This will not work, for example, if only some lines have
            # an idx specified (in some cases)
            for line in open(filelist, mode="rb"):
                r = json.loads(line)
                input_line_idx = r.get("idx")
                if input_line_idx is not None:
                    # We only have input indices if we have a transient retry.
                    is_transient_retry = True
                else:
                    input_line_idx = line_idx
                line_idx += 1
                local = r["local"]
                url = r["url"]
                content_type = r.get("content_type", None)
                metadata = r.get("metadata", None)
                encryption = r.get("encryption", None)
                if not os.path.exists(local):
                    exit(ERROR_LOCAL_FILE_NOT_FOUND, local)
                yield input_line_idx, local, url, content_type, metadata, encryption

    def _make_url(idx, local, user_url, content_type, metadata, encryption):
        src = urlparse(user_url)
        url = S3Url(
            url=user_url,
            bucket=src.netloc,
            path=src.path.lstrip("/"),
            local=local,
            prefix=None,
            content_type=content_type,
            metadata=metadata,
            idx=idx,
            encryption=encryption,
        )
        if src.scheme != "s3":
            exit(ERROR_INVALID_URL, url)
        if not src.path:
            exit(ERROR_NOT_FULL_PATH, url)
        return url

    s3config = S3Config(
        s3role,
        json.loads(s3sessionvars) if s3sessionvars else None,
        json.loads(s3clientparams) if s3clientparams else None,
    )

    urls = list(starmap(_make_url, _files()))
    ul_op = "upload"
    if not overwrite:
        ul_op = "info_upload"
    sz_results = process_urls(
        ul_op, urls, verbose, inject_failure, num_workers, s3config
    )
    retry_lines = []
    out_lines = []
    denied_url = None
    for url, sz in zip(urls, sz_results):
        # sz is None if the file wasn't uploaded (no overwrite), 0 if uploaded OK
        # or the error code if not (error code here will only be
        # ERROR_TRANSIENT or ERROR_URL_ACCESS_DENIED
        if sz is None:
            if listing:
                # We keep a position for it in our out list in case of retries
                out_lines.append("%d %s\n" % (url.idx, TRANSIENT_RETRY_LINE_CONTENT))
            continue
        elif listing and sz == 0:
            out_lines.append(format_result_line(url.idx, url.url) + "\n")
        elif sz == -ERROR_TRANSIENT:
            retry_lines.append(
                json.dumps(
                    {
                        "idx": url.idx,
                        "url": url.url,
                        "local": url.local,
                        "content_type": url.content_type,
                        "metadata": url.metadata,
                        "encryption": url.encryption,
                    }
                )
                + "\n"
            )
            # Output something to get a total count the first time around
            if not is_transient_retry:
                out_lines.append("%d %s\n" % (url.idx, TRANSIENT_RETRY_LINE_CONTENT))
        elif sz == -ERROR_URL_ACCESS_DENIED:
            # We do NOT break because we want to be able to accurately report all
            # the files uploaded after retries.
            denied_url = url
    if denied_url is not None:
        exit(ERROR_URL_ACCESS_DENIED, denied_url)

    if out_lines:
        sys.stdout.writelines(out_lines)
        sys.stdout.flush()

    if retry_lines:
        sys.stderr.write("%s\n" % TRANSIENT_RETRY_START_LINE)
        sys.stderr.writelines(retry_lines)
        sys.stderr.flush()
        sys.exit(ERROR_TRANSIENT)


def _populate_prefixes(prefixes, inputs):
    # Returns a tuple: first element is the prefix index, the second element is the
    # prefix and the third element is the optional range (or None if the entire prefix
    # is requested).
    # We again assume that the indices, if provided, are correct. This is again only
    # used for the transient error retry so users should not use this directly.
    is_transient_retry = False
    if prefixes:
        prefixes = [(idx, url_unquote(p), None) for idx, p in enumerate(prefixes)]
    else:
        prefixes = []
    if inputs:
        with open(inputs, mode="rb") as f:
            for idx, l in enumerate(f, start=len(prefixes)):
                s = l.split(b" ")
                if len(s) == 1:
                    url = url_unquote(s[0].strip())
                    prefixes.append((idx, url, url, None))
                elif len(s) == 2:
                    url = url_unquote(s[0].strip())
                    prefixes.append((idx, url, url, url_unquote(s[1].strip())))
                else:
                    is_transient_retry = True
                    if len(s) == 3:
                        prefix = url = url_unquote(s[1].strip())
                        range_info = url_unquote(s[2].strip())
                    else:
                        # Special case when we have both prefix and URL -- this is
                        # used in recursive gets for example
                        prefix = url_unquote(s[1].strip())
                        url = url_unquote(s[2].strip())
                        range_info = url_unquote(s[3].strip())
                    if range_info == "<norange>":
                        range_info = None
                    prefixes.append(
                        (int(url_unquote(s[0].strip())), prefix, url, range_info)
                    )
    return prefixes, is_transient_retry


@cli.command(help="Download files from S3")
@tracing.cli("s3op/get")
@click.option(
    "--recursive/--no-recursive",
    default=False,
    show_default=True,
    help="Download prefixes recursively.",
)
@click.option(
    "--verify/--no-verify",
    default=True,
    show_default=True,
    help="Verify that files were loaded correctly.",
)
@click.option(
    "--info/--no-info",
    default=True,
    show_default=True,
    help="Return user tags and content-type",
)
@click.option(
    "--allow-missing/--no-allow-missing",
    default=False,
    show_default=True,
    help="Do not exit if missing files are detected. Implies --verify.",
)
@common_options
@non_lst_common_options
@click.argument("prefixes", nargs=-1)
def get(
    prefixes,
    recursive=None,
    num_workers=None,
    inputs=None,
    verify=None,
    info=None,
    allow_missing=None,
    verbose=None,
    listing=None,
    s3role=None,
    s3sessionvars=None,
    s3clientparams=None,
    inject_failure=0,
):

    s3config = S3Config(
        s3role,
        json.loads(s3sessionvars) if s3sessionvars else None,
        json.loads(s3clientparams) if s3clientparams else None,
    )

    # Construct a list of URL (prefix) objects
    urllist = []
    to_iterate, is_transient_retry = _populate_prefixes(prefixes, inputs)
    for idx, prefix, url, r in to_iterate:
        src = urlparse(url)
        url = S3Url(
            url=url,
            bucket=src.netloc,
            path=src.path.lstrip("/"),
            local=generate_local_path(url, range=r),
            prefix=prefix,
            range=r,
            idx=idx,
        )
        if src.scheme != "s3":
            exit(ERROR_INVALID_URL, url)
        if not recursive and not src.path:
            exit(ERROR_NOT_FULL_PATH, url)
        urllist.append(url)
    # Construct a URL->size mapping and get content-type and metadata if needed
    op = None
    dl_op = "download"
    if recursive:
        op = partial(op_list_prefix, s3config)
    if verify or verbose or info:
        dl_op = "info_download"
    if op:
        if is_transient_retry:
            raise RuntimeError("--recursive not allowed for transient retries")
        urls = []
        # NOTE - we must retain the order of prefixes requested
        # and the listing order returned by S3
        for success, prefix_url, ret in parallel_op(op, urllist, num_workers):
            if success:
                urls.extend(ret)
            elif ret == ERROR_URL_NOT_FOUND and allow_missing:
                urls.append((prefix_url, None))
            else:
                exit(ret, prefix_url)
        # We re-index here since we may have pulled in a bunch more stuff. On a transient
        # retry, we never have recursive so we would not re-index
        for idx, (url, _) in enumerate(urls):
            url.idx = idx
    else:
        # pretend zero size since we don't need it for anything.
        # it can't be None though, to make sure the listing below
        # works correctly (None denotes a missing file)
        urls = [(prefix_url, 0) for prefix_url in urllist]

    # exclude the non-existent files from loading
    to_load = [url for url, size in urls if size is not None]
    sz_results = process_urls(
        dl_op, to_load, verbose, inject_failure, num_workers, s3config
    )
    # We check if there is any access denied
    retry_lines = []
    out_lines = []
    denied_url = None
    missing_url = None
    verify_info = []
    idx_in_sz = 0
    for url, _ in urls:
        sz = None
        # to_load contains an ordered subset of urls
        if idx_in_sz != len(to_load) and url.url == to_load[idx_in_sz].url:
            sz = sz_results[idx_in_sz]
            idx_in_sz += 1
        if listing and sz is None:
            out_lines.append(format_result_line(url.idx, url.url) + "\n")
        elif listing and sz >= 0:
            out_lines.append(
                format_result_line(url.idx, url.prefix, url.url, url.local) + "\n"
            )
            if verify:
                verify_info.append((url, sz))
        elif sz == -ERROR_OUT_OF_DISK_SPACE:
            exit(ERROR_OUT_OF_DISK_SPACE, url)
        elif sz == -ERROR_URL_ACCESS_DENIED:
            denied_url = url
            break
        elif sz == -ERROR_URL_NOT_FOUND:
            if missing_url is None:
                missing_url = url
            if not allow_missing:
                break
            out_lines.append(format_result_line(url.idx, url.url) + "\n")
        elif sz == -ERROR_TRANSIENT:
            retry_lines.append(
                " ".join(
                    [
                        str(url.idx),
                        url_quote(url.prefix).decode(encoding="utf-8"),
                        url_quote(url.url).decode(encoding="utf-8"),
                        (
                            url_quote(url.range).decode(encoding="utf-8")
                            if url.range
                            else "<norange>"
                        ),
                    ]
                )
                + "\n"
            )
            # First time around, we output something to indicate the total length
            if not is_transient_retry:
                out_lines.append("%d %s\n" % (url.idx, TRANSIENT_RETRY_LINE_CONTENT))

    if denied_url is not None:
        exit(ERROR_URL_ACCESS_DENIED, denied_url)

    if not allow_missing and missing_url is not None:
        exit(ERROR_URL_NOT_FOUND, missing_url)

    # Postprocess
    if verify:
        verify_results(verify_info, verbose=verbose)

    if out_lines:
        sys.stdout.writelines(out_lines)
        sys.stdout.flush()

    if retry_lines:
        sys.stderr.write("%s\n" % TRANSIENT_RETRY_START_LINE)
        sys.stderr.writelines(retry_lines)
        sys.stderr.flush()
        sys.exit(ERROR_TRANSIENT)


@cli.command(help="Get info about files from S3")
@common_options
@non_lst_common_options
@click.argument("prefixes", nargs=-1)
def info(
    prefixes,
    num_workers=None,
    inputs=None,
    verbose=None,
    listing=None,
    s3role=None,
    s3sessionvars=None,
    s3clientparams=None,
    inject_failure=0,
):

    s3config = S3Config(
        s3role,
        json.loads(s3sessionvars) if s3sessionvars else None,
        json.loads(s3clientparams) if s3clientparams else None,
    )

    # Construct a list of URL (prefix) objects
    urllist = []
    to_iterate, is_transient_retry = _populate_prefixes(prefixes, inputs)
    for idx, prefix, url, _ in to_iterate:
        src = urlparse(url)
        url = S3Url(
            url=url,
            bucket=src.netloc,
            path=src.path.lstrip("/"),
            local=generate_local_path(url, suffix="info"),
            prefix=prefix,
            range=None,
            idx=idx,
        )
        if src.scheme != "s3":
            exit(ERROR_INVALID_URL, url)
        urllist.append(url)

    sz_results = process_urls(
        "info", urllist, verbose, inject_failure, num_workers, s3config
    )

    retry_lines = []
    out_lines = []
    for idx, sz in enumerate(sz_results):
        url = urllist[idx]
        if listing and sz != -ERROR_TRANSIENT:
            out_lines.append(
                format_result_line(url.idx, url.prefix, url.url, url.local) + "\n"
            )
        else:
            retry_lines.append(
                "%d %s <norange>\n"
                % (url.idx, url_quote(url.url).decode(encoding="utf-8"))
            )
            if not is_transient_retry:
                out_lines.append("%d %s\n" % (url.idx, TRANSIENT_RETRY_LINE_CONTENT))

    if out_lines:
        sys.stdout.writelines(out_lines)
        sys.stdout.flush()

    if retry_lines:
        sys.stderr.write("%s\n" % TRANSIENT_RETRY_START_LINE)
        sys.stderr.writelines(retry_lines)
        sys.stderr.flush()
        sys.exit(ERROR_TRANSIENT)


if __name__ == "__main__":
    cli(auto_envvar_prefix="S3OP")



from metaflow.decorators import FlowDecorator
from metaflow.exception import MetaflowException


class ExitHookDecorator(FlowDecorator):
    name = "exit_hook"
    allow_multiple = True

    defaults = {
        "on_success": [],
        "on_error": [],
        "options": {},
    }

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        on_success = self.attributes["on_success"]
        on_error = self.attributes["on_error"]

        if not on_success and not on_error:
            raise MetaflowException(
                "Choose at least one of the options on_success/on_error"
            )

        self.success_hooks = []
        self.error_hooks = []
        for success_fn in on_success:
            if isinstance(success_fn, str):
                self.success_hooks.append(success_fn)
            elif callable(success_fn):
                self.success_hooks.append(success_fn.__name__)
            else:
                raise ValueError(
                    "Exit hooks inside 'on_success' must be a function or a string referring to the function"
                )

        for error_fn in on_error:
            if isinstance(error_fn, str):
                self.error_hooks.append(error_fn)
            elif callable(error_fn):
                self.error_hooks.append(error_fn.__name__)
            else:
                raise ValueError(
                    "Exit hooks inside 'on_error' must be a function or a string referring to the function"
                )



import collections
import glob
import json
import os
import re
import random
import tempfile
import time
from collections import namedtuple
from typing import List

from metaflow.exception import MetaflowInternalError, MetaflowTaggingError
from metaflow.metadata_provider.metadata import ObjectOrder
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.metadata_provider import MetadataProvider
from metaflow.tagging_util import MAX_USER_TAG_SET_SIZE, validate_tags


class LocalMetadataProvider(MetadataProvider):
    TYPE = "local"

    def __init__(self, environment, flow, event_logger, monitor):
        super(LocalMetadataProvider, self).__init__(
            environment, flow, event_logger, monitor
        )

    @classmethod
    def compute_info(cls, val):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        v = os.path.realpath(os.path.join(val, DATASTORE_LOCAL_DIR))
        if os.path.isdir(v):
            LocalStorage.datastore_root = v
            return val
        raise ValueError(
            "Could not find directory %s in directory %s" % (DATASTORE_LOCAL_DIR, val)
        )

    @classmethod
    def default_info(cls):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        def print_clean(line, **kwargs):
            print(line)

        v = LocalStorage.get_datastore_root_from_config(
            print_clean, create_on_absent=False
        )
        if v is None:
            return (
                "<No %s directory found in current working tree>" % DATASTORE_LOCAL_DIR
            )
        return os.path.dirname(v)

    def version(self):
        return "local"

    def new_run_id(self, tags=None, sys_tags=None):
        # We currently just use the timestamp to create an ID. We can be reasonably certain
        # that it is unique and this makes it possible to do without coordination or
        # reliance on POSIX locks in the filesystem.
        run_id = "%d" % (time.time() * 1e6)
        self._new_run(run_id, tags, sys_tags)
        return run_id

    def register_run_id(self, run_id, tags=None, sys_tags=None):
        try:
            # This metadata provider only generates integer IDs so if this is
            # an integer, we don't register it again (since it was "registered"
            # on creation). However, some IDs are created outside the metadata
            # provider and need to be properly registered
            int(run_id)
            return False
        except ValueError:
            return self._new_run(run_id, tags, sys_tags)

    def new_task_id(self, run_id, step_name, tags=None, sys_tags=None):
        self._task_id_seq += 1
        task_id = str(self._task_id_seq)
        self._new_task(run_id, step_name, task_id, tags=tags, sys_tags=sys_tags)
        return task_id

    def register_task_id(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        try:
            # Same logic as register_run_id
            int(task_id)
        except ValueError:
            return self._new_task(
                run_id,
                step_name,
                task_id,
                attempt=attempt,
                tags=tags,
                sys_tags=sys_tags,
            )
        else:
            self._register_system_metadata(run_id, step_name, task_id, attempt)
            return False

    def register_data_artifacts(
        self, run_id, step_name, task_id, attempt_id, artifacts
    ):
        meta_dir = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        artlist = self._artifacts_to_json(
            run_id, step_name, task_id, attempt_id, artifacts
        )
        artdict = {"%d_artifact_%s" % (attempt_id, art["name"]): art for art in artlist}
        self._save_meta(meta_dir, artdict)

    def register_metadata(self, run_id, step_name, task_id, metadata):
        meta_dir = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        metalist = self._metadata_to_json(run_id, step_name, task_id, metadata)
        ts = int(round(time.time() * 1000))
        metadict = {
            "sysmeta_%s_%d" % (meta["field_name"], ts): meta for meta in metalist
        }
        self._save_meta(meta_dir, metadict)

    @classmethod
    def _mutate_user_tags_for_run(
        cls, flow_id, run_id, tags_to_add=None, tags_to_remove=None
    ):
        MutationResult = namedtuple(
            "MutationResult", field_names="tags_are_consistent tags"
        )

        def _optimistically_mutate():
            # get existing tags
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, flow_id, run_id
            )
            if not run:
                raise MetaflowTaggingError(
                    msg="Run not found (%s, %s)" % (flow_id, run_id)
                )
            existing_user_tag_set = frozenset(run["tags"])
            existing_system_tag_set = frozenset(run["system_tags"])
            tags_to_remove_set = frozenset(tags_to_remove)
            # make sure no existing system tags get added as a user tag
            tags_to_add_set = frozenset(tags_to_add) - existing_system_tag_set

            # from this point on we work with sets of tags only

            if tags_to_remove_set & existing_system_tag_set:
                raise MetaflowTaggingError(
                    msg="Cannot remove a tag that is an existing system tag (%s)"
                    % str(sorted(tags_to_remove_set & existing_system_tag_set))
                )

            # remove tags first, then add
            next_user_tags_set = (
                existing_user_tag_set - tags_to_remove_set
            ) | tags_to_add_set
            # we think it will be a no-op, so let's return right away
            if next_user_tags_set == existing_user_tag_set:
                return MutationResult(
                    tags=next_user_tags_set,
                    tags_are_consistent=True,
                )

            validate_tags(next_user_tags_set, existing_tags=existing_user_tag_set)

            # write new tag set to file system
            LocalMetadataProvider._persist_tags_for_run(
                flow_id, run_id, next_user_tags_set, existing_system_tag_set
            )

            # read tags back from file system to see if our optimism is misplaced
            # I.e. did a concurrent mutate overwrite our change
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, flow_id, run_id
            )
            if not run:
                raise MetaflowTaggingError(
                    msg="Run not found for read-back check (%s, %s)" % (flow_id, run_id)
                )
            final_tag_set = frozenset(run["tags"])
            if tags_to_add_set - final_tag_set:
                return MutationResult(tags=final_tag_set, tags_are_consistent=False)
            if (
                tags_to_remove_set & final_tag_set
            ) - tags_to_add_set:  # Remove before add, remember?  Account for this
                return MutationResult(tags=final_tag_set, tags_are_consistent=False)

            return MutationResult(tags=final_tag_set, tags_are_consistent=True)

        tries = 1
        # try up to 5 times, with a gentle exponential backoff (1.1-1.3x)
        while True:
            mutation_result = _optimistically_mutate()
            if mutation_result.tags_are_consistent:
                return mutation_result.tags
            if tries >= 5:
                break
            time.sleep(0.3 * random.uniform(1.1, 1.3) ** tries)
            tries += 1
        raise MetaflowTaggingError(
            "Tagging failed due to too many conflicting updates from other processes"
        )

    @classmethod
    def filter_tasks_by_metadata(
        cls,
        flow_name: str,
        run_id: str,
        step_name: str,
        field_name: str,
        pattern: str,
    ) -> List[str]:
        """
        Filter tasks by metadata field and pattern, returning task pathspecs that match criteria.

        Parameters
        ----------
        flow_name : str
            Identifier for the flow
        run_id : str
            Identifier for the run
        step_name : str
            Name of the step to query tasks from
        field_name : str
            Name of metadata field to query
        pattern : str
            Pattern to match in metadata field value

        Returns
        -------
        List[str]
            List of task pathspecs that match the query criteria
        """
        tasks = cls.get_object("step", "task", {}, None, flow_name, run_id, step_name)
        if not tasks:
            return []

        regex = re.compile(pattern)
        matching_task_pathspecs = []

        for task in tasks:
            task_id = task.get("task_id")
            if not task_id:
                continue

            if pattern == ".*":
                # If the pattern is ".*", we can match all tasks without reading metadata
                matching_task_pathspecs.append(
                    f"{flow_name}/{run_id}/{step_name}/{task_id}"
                )
                continue

            metadata = cls.get_object(
                "task", "metadata", {}, None, flow_name, run_id, step_name, task_id
            )

            if any(
                meta.get("field_name") == field_name
                and regex.match(meta.get("value", ""))
                for meta in metadata
            ):
                matching_task_pathspecs.append(
                    f"{flow_name}/{run_id}/{step_name}/{task_id}"
                )

        return matching_task_pathspecs

    @classmethod
    def _get_object_internal(
        cls, obj_type, obj_order, sub_type, sub_order, filters, attempt, *args
    ):
        # This is guaranteed by MetaflowProvider.get_object(), sole intended caller
        if obj_type in ("metadata", "self"):
            raise MetaflowInternalError(msg="Type %s is not allowed" % obj_type)

        if obj_type not in ("root", "flow", "run", "step", "task", "artifact"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

        from metaflow.plugins.datastores.local_storage import LocalStorage

        if obj_type == "artifact":
            # Artifacts are actually part of the tasks in the filesystem
            # E.g. we get here for (obj_type, sub_type) == (artifact, self)
            obj_type = "task"
            sub_type = "artifact"
            sub_order = obj_order
            obj_order = obj_order - 1

        if obj_type != ObjectOrder.order_to_type(obj_order):
            raise MetaflowInternalError(
                "Object type order mismatch %s %s"
                % (obj_type, ObjectOrder.order_to_type(obj_order))
            )
        if sub_type != ObjectOrder.order_to_type(sub_order):
            raise MetaflowInternalError(
                "Sub type order mismatch %s %s"
                % (sub_type, ObjectOrder.order_to_type(sub_order))
            )

        RUN_ORDER = ObjectOrder.type_to_order("run")

        if obj_type not in ("root", "flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

        # Special handling of self, artifact, and metadata
        if sub_type == "self":
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return None
            self_file = os.path.join(meta_path, "_self.json")
            if os.path.isfile(self_file):
                obj = MetadataProvider._apply_filter(
                    [LocalMetadataProvider._read_json_file(self_file)], filters
                )[0]
                # For non-descendants of a run, we are done

                if obj_order <= RUN_ORDER:
                    return obj

                if obj_type not in ("step", "task"):
                    raise MetaflowInternalError(
                        msg="Unexpected object type %s" % obj_type
                    )
                run = LocalMetadataProvider.get_object(
                    "run", "self", {}, None, *args[:RUN_ORDER]  # *[flow_id, run_id]
                )
                if not run:
                    raise MetaflowInternalError(
                        msg="Could not find run %s" % str(args[:RUN_ORDER])
                    )

                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])
                return obj
            return None

        if sub_type == "artifact":
            if obj_type not in ("root", "flow", "run", "step", "task"):
                raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            result = []
            if meta_path is None:
                return result

            successful_attempt = attempt
            if successful_attempt is None:
                attempt_done_files = os.path.join(meta_path, "sysmeta_attempt-done_*")
                attempts_done = sorted(glob.iglob(attempt_done_files))
                if attempts_done:
                    successful_attempt = int(
                        LocalMetadataProvider._read_json_file(attempts_done[-1])[
                            "value"
                        ]
                    )
            if successful_attempt is not None:
                which_artifact = "*"
                if len(args) >= sub_order:
                    which_artifact = args[sub_order - 1]
                artifact_files = os.path.join(
                    meta_path,
                    "%d_artifact_%s.json" % (successful_attempt, which_artifact),
                )
                for obj in glob.iglob(artifact_files):
                    result.append(LocalMetadataProvider._read_json_file(obj))

            # We are getting artifacts. We should overlay with ancestral run's tags
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, *args[:RUN_ORDER]  # *[flow_id, run_id]
            )
            if not run:
                raise MetaflowInternalError(
                    msg="Could not find run %s" % str(args[:RUN_ORDER])
                )
            for obj in result:
                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])

            if len(result) == 1:
                return result[0]
            return result

        if sub_type == "metadata":
            # artifact is not expected because if obj_type=artifact on function entry, we transform to =task
            if obj_type not in ("root", "flow", "run", "step", "task"):
                raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)
            result = []
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return result
            files = os.path.join(meta_path, "sysmeta_*")
            for obj in glob.iglob(files):
                result.append(LocalMetadataProvider._read_json_file(obj))
            return result

        # For the other types, we locate all the objects we need to find and return them
        if obj_type not in ("root", "flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)
        if sub_type not in ("flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="unexpected sub type %s" % sub_type)
        obj_path = LocalMetadataProvider._make_path(
            *args[:obj_order], create_on_absent=False
        )
        result = []
        if obj_path is None:
            return result
        skip_dirs = "*/" * (sub_order - obj_order)
        all_meta = os.path.join(obj_path, skip_dirs, LocalStorage.METADATA_DIR)
        SelfInfo = collections.namedtuple("SelfInfo", ["filepath", "run_id"])
        self_infos = []
        for meta_path in glob.iglob(all_meta):
            self_file = os.path.join(meta_path, "_self.json")
            if not os.path.isfile(self_file):
                continue
            run_id = None
            # flow and run do not need info from ancestral run
            if sub_type in ("step", "task"):
                run_id = LocalMetadataProvider._deduce_run_id_from_meta_dir(
                    meta_path, sub_type
                )
                # obj_type IS run, or more granular than run, let's do sanity check vs args
                if obj_order >= RUN_ORDER:
                    if run_id != args[RUN_ORDER - 1]:
                        raise MetaflowInternalError(
                            msg="Unexpected run id %s deduced from meta path" % run_id
                        )
            self_infos.append(SelfInfo(filepath=self_file, run_id=run_id))

        for self_info in self_infos:
            obj = LocalMetadataProvider._read_json_file(self_info.filepath)
            if self_info.run_id:
                flow_id_from_args = args[0]
                run = LocalMetadataProvider.get_object(
                    "run",
                    "self",
                    {},
                    None,
                    flow_id_from_args,
                    self_info.run_id,
                )
                if not run:
                    raise MetaflowInternalError(
                        msg="Could not find run %s, %s"
                        % (flow_id_from_args, self_info.run_id)
                    )
                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])
            result.append(obj)

        return MetadataProvider._apply_filter(result, filters)

    @staticmethod
    def _deduce_run_id_from_meta_dir(meta_dir_path, sub_type):
        curr_order = ObjectOrder.type_to_order(sub_type)
        levels_to_ascend = curr_order - ObjectOrder.type_to_order("run")
        if levels_to_ascend < 0:
            return None
        curr_path = meta_dir_path
        for _ in range(levels_to_ascend + 1):  # +1 to account for ../_meta
            curr_path, _ = os.path.split(curr_path)
        _, run_id = os.path.split(curr_path)
        if not run_id:
            raise MetaflowInternalError(
                "Failed to deduce run_id from meta dir %s" % meta_dir_path
            )
        return run_id

    @staticmethod
    def _makedirs(path):
        # this is for python2 compatibility.
        # Python3 has os.makedirs(exist_ok=True).
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                # Error raised when directory exists
                return
            else:
                raise

    @staticmethod
    def _persist_tags_for_run(flow_id, run_id, tags, system_tags):
        subpath = LocalMetadataProvider._create_and_get_metadir(
            flow_name=flow_id, run_id=run_id
        )
        selfname = os.path.join(subpath, "_self.json")
        if not os.path.isfile(selfname):
            raise MetaflowInternalError(
                msg="Could not verify Run existence on disk - missing %s" % selfname
            )
        LocalMetadataProvider._save_meta(
            subpath,
            {
                "_self": MetadataProvider._run_to_json_static(
                    flow_id, run_id=run_id, tags=tags, sys_tags=system_tags
                )
            },
            allow_overwrite=True,
        )

    def _ensure_meta(
        self, obj_type, run_id, step_name, task_id, tags=None, sys_tags=None
    ):
        if tags is None:
            tags = set()
        if sys_tags is None:
            sys_tags = set()
        subpath = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        selfname = os.path.join(subpath, "_self.json")
        self._makedirs(subpath)
        if os.path.isfile(selfname):
            # There is a race here, but we are not aiming to make this as solid as
            # the metadata service. This is used primarily for concurrent resumes,
            # so it is highly unlikely that this combination (multiple resumes of
            # the same flow on the same machine) happens.
            return False
        # In this case the metadata information does not exist, so we create it
        self._save_meta(
            subpath,
            {
                "_self": self._object_to_json(
                    obj_type,
                    run_id,
                    step_name,
                    task_id,
                    self.sticky_tags.union(tags),
                    self.sticky_sys_tags.union(sys_tags),
                )
            },
        )
        return True

    def _new_run(self, run_id, tags=None, sys_tags=None):
        self._ensure_meta("flow", None, None, None)
        return self._ensure_meta("run", run_id, None, None, tags, sys_tags)

    def _new_task(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        self._ensure_meta("step", run_id, step_name, None)
        to_return = self._ensure_meta(
            "task", run_id, step_name, task_id, tags, sys_tags
        )
        self._register_system_metadata(run_id, step_name, task_id, attempt)
        return to_return

    @staticmethod
    def _make_path(
        flow_name=None, run_id=None, step_name=None, task_id=None, create_on_absent=True
    ):

        from metaflow.plugins.datastores.local_storage import LocalStorage

        if LocalStorage.datastore_root is None:

            def print_clean(line, **kwargs):
                print(line)

            LocalStorage.datastore_root = LocalStorage.get_datastore_root_from_config(
                print_clean, create_on_absent=create_on_absent
            )
        if LocalStorage.datastore_root is None:
            return None

        if flow_name is None:
            return LocalStorage.datastore_root
        components = []
        if flow_name:
            components.append(flow_name)
            if run_id:
                components.append(run_id)
                if step_name:
                    components.append(step_name)
                    if task_id:
                        components.append(task_id)
        return LocalStorage().full_uri(LocalStorage.path_join(*components))

    @staticmethod
    def _create_and_get_metadir(
        flow_name=None, run_id=None, step_name=None, task_id=None
    ):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        root_path = LocalMetadataProvider._make_path(
            flow_name, run_id, step_name, task_id
        )
        subpath = os.path.join(root_path, LocalStorage.METADATA_DIR)
        LocalMetadataProvider._makedirs(subpath)
        return subpath

    @staticmethod
    def _get_metadir(flow_name=None, run_id=None, step_name=None, task_id=None):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        root_path = LocalMetadataProvider._make_path(
            flow_name, run_id, step_name, task_id, create_on_absent=False
        )
        if root_path is None:
            return None
        subpath = os.path.join(root_path, LocalStorage.METADATA_DIR)
        if os.path.isdir(subpath):
            return subpath
        return None

    @staticmethod
    def _dump_json_to_file(filepath, data, allow_overwrite=False):
        if os.path.isfile(filepath) and not allow_overwrite:
            return
        try:
            with tempfile.NamedTemporaryFile(
                mode="w", dir=os.path.dirname(filepath), delete=False
            ) as f:
                json.dump(data, f)
            os.rename(f.name, filepath)
        finally:
            # clean up in case anything goes wrong
            if f and os.path.isfile(f.name):
                os.remove(f.name)

    @staticmethod
    def _read_json_file(filepath):
        with open(filepath, "r") as f:
            return json.load(f)

    @staticmethod
    def _save_meta(root_dir, metadict, allow_overwrite=False):
        for name, datum in metadict.items():
            filename = os.path.join(root_dir, "%s.json" % name)
            LocalMetadataProvider._dump_json_to_file(
                filename, datum, allow_overwrite=allow_overwrite
            )



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from ._functions import TypeCheckFailCallback

T = TypeVar("T")


class ForwardRefPolicy(Enum):
    """
    Defines how unresolved forward references are handled.

    Members:

    * ``ERROR``: propagate the :exc:`NameError` when the forward reference lookup fails
    * ``WARN``: emit a :class:`~.TypeHintWarning` if the forward reference lookup fails
    * ``IGNORE``: silently skip checks for unresolveable forward references
    """

    ERROR = auto()
    WARN = auto()
    IGNORE = auto()


class CollectionCheckStrategy(Enum):
    """
    Specifies how thoroughly the contents of collections are type checked.

    This has an effect on the following built-in checkers:

    * ``AbstractSet``
    * ``Dict``
    * ``List``
    * ``Mapping``
    * ``Set``
    * ``Tuple[<type>, ...]`` (arbitrarily sized tuples)

    Members:

    * ``FIRST_ITEM``: check only the first item
    * ``ALL_ITEMS``: check all items
    """

    FIRST_ITEM = auto()
    ALL_ITEMS = auto()

    def iterate_samples(self, collection: Collection[T]) -> Collection[T]:
        if self is CollectionCheckStrategy.FIRST_ITEM:
            if len(collection):
                return [next(iter(collection))]
            else:
                return ()
        else:
            return collection


@dataclass
class TypeCheckConfiguration:
    """
     You can change Typeguard's behavior with these settings.

    .. attribute:: typecheck_fail_callback
       :type: Callable[[TypeCheckError, TypeCheckMemo], Any]

         Callable that is called when type checking fails.

         Default: ``None`` (the :exc:`~.TypeCheckError` is raised directly)

    .. attribute:: forward_ref_policy
       :type: ForwardRefPolicy

         Specifies what to do when a forward reference fails to resolve.

         Default: ``WARN``

    .. attribute:: collection_check_strategy
       :type: CollectionCheckStrategy

         Specifies how thoroughly the contents of collections (list, dict, etc.) are
         type checked.

         Default: ``FIRST_ITEM``

    .. attribute:: debug_instrumentation
       :type: bool

         If set to ``True``, the code of modules or functions instrumented by typeguard
         is printed to ``sys.stderr`` after the instrumentation is done

         Requires Python 3.9 or newer.

         Default: ``False``
    """

    forward_ref_policy: ForwardRefPolicy = ForwardRefPolicy.WARN
    typecheck_fail_callback: TypeCheckFailCallback | None = None
    collection_check_strategy: CollectionCheckStrategy = (
        CollectionCheckStrategy.FIRST_ITEM
    )
    debug_instrumentation: bool = False


global_config = TypeCheckConfiguration()



"""
Welcome to Metaflow!

Metaflow is a microframework for data science projects.

There are two main use cases for this package:

1) You can define new flows using the `FlowSpec`
   class and related decorators.

2) You can access and inspect existing flows.
   You can instantiate a `Metaflow` class to
   get an entry point to all existing objects.

# How to work with flows

A flow is a directed graph of Python functions called steps.
Metaflow takes care of executing these steps one by one in various
environments, such as on a local laptop or compute environments
(such as AWS Batch for example). It snapshots
data and code related to each run, so you can resume, reproduce,
and inspect results easily at a later point in time.

Here is a high-level overview of objects related to flows:

 [ FlowSpec ]     (0) Base class for flows.
 [ MyFlow ]       (1) Subclass from FlowSpec to define a new flow.

define new flows
----------------- (2) Run MyFlow on the command line.
access results

 [ Flow ]         (3) Access your flow with `Flow('MyFlow')`.
 [ Run ]          (4) Access a specific run with `Run('MyFlow/RunID')`.
 [ Step ]         (5) Access a specific step by its name, e.g. `run['end']`.
 [ Task ]         (6) Access a task related to the step with `step.task`.
 [ DataArtifact ] (7) Access data of a task with `task.data`.

# More questions?

If you have any questions, feel free to post a bug report/question on the
Metaflow GitHub page.
"""

import os
import sys

from metaflow.extension_support import (
    alias_submodules,
    get_modules,
    lazy_load_aliases,
    load_globals,
    load_module,
    EXT_PKG,
    _ext_debug,
)

# We load the module overrides *first* explicitly. Non overrides can be loaded
# in toplevel as well but these can be loaded first if needed. Note that those
# modules should be careful not to include anything in Metaflow at their top-level
# as it is likely to not work.
_override_modules = []
_tl_modules = []
try:
    _modules_to_import = get_modules("toplevel")

    for m in _modules_to_import:
        override_module = m.module.__dict__.get("module_overrides", None)
        if override_module is not None:
            _override_modules.append(
                ".".join([EXT_PKG, m.tl_package, "toplevel", override_module])
            )
        tl_module = m.module.__dict__.get("toplevel", None)
        if tl_module is not None:
            _tl_modules.append(
                (
                    m.package_name,
                    ".".join([EXT_PKG, m.tl_package, "toplevel", tl_module]),
                )
            )
    _ext_debug("Got overrides to load: %s" % _override_modules)
    _ext_debug("Got top-level imports: %s" % str(_tl_modules))
except Exception as e:
    _ext_debug("Error in importing toplevel/overrides: %s" % e)

# Load overrides now that we have them (in the proper order)
for m in _override_modules:
    extension_module = load_module(m)
    if extension_module:
        # We load only modules
        tl_package = m.split(".")[1]
        lazy_load_aliases(alias_submodules(extension_module, tl_package, None))

# Utilities
from .multicore_utils import parallel_imap_unordered, parallel_map
from .metaflow_profile import profile

# current runtime singleton
from .metaflow_current import current

# Flow spec
from .flowspec import FlowSpec

from .parameters import Parameter, JSONTypeClass, JSONType

from .user_configs.config_parameters import Config, ConfigValue, config_expr
from .user_decorators.user_step_decorator import (
    UserStepDecorator,
    StepMutator,
    user_step_decorator,
    USER_SKIP_STEP,
)
from .user_decorators.user_flow_decorator import FlowMutator

# data layer
# For historical reasons, we make metaflow.plugins.datatools accessible as
# metaflow.datatools. S3 is also a tool that has historically been available at the
# top-level so keep as is.
lazy_load_aliases({"metaflow.datatools": "metaflow.plugins.datatools"})
from .plugins.datatools import S3

# includefile
from .includefile import IncludeFile

# Decorators
from .decorators import step, _import_plugin_decorators


# Parsers (for configs) for now
from .plugins import _import_tl_plugins

_import_tl_plugins(globals())

# this auto-generates decorator functions from Decorator objects
# in the top-level metaflow namespace
_import_plugin_decorators(globals())
# Setting card import for only python 3.6
if sys.version_info[0] >= 3 and sys.version_info[1] >= 6:
    from . import cards

# Client
from .client import (
    namespace,
    get_namespace,
    default_namespace,
    metadata,
    get_metadata,
    default_metadata,
    Metaflow,
    Flow,
    Run,
    Step,
    Task,
    DataArtifact,
)

# Import data class within tuple_util but not introduce new symbols.
from . import tuple_util

# Runner API
if sys.version_info >= (3, 7):
    from .runner.metaflow_runner import Runner
    from .runner.nbrun import NBRunner
    from .runner.deployer import Deployer
    from .runner.deployer import DeployedFlow
    from .runner.nbdeploy import NBDeployer

__ext_tl_modules__ = []
_ext_debug("Loading top-level modules")
for pkg_name, m in _tl_modules:
    extension_module = load_module(m)
    if extension_module:
        tl_package = m.split(".")[1]
        load_globals(extension_module, globals(), extra_indent=True)
        lazy_load_aliases(
            alias_submodules(extension_module, tl_package, None, extra_indent=True)
        )
        __ext_tl_modules__.append((pkg_name, extension_module))

# Erase all temporary names to avoid leaking things
for _n in [
    "_ext_debug",
    "alias_submodules",
    "get_modules",
    "lazy_load_aliases",
    "load_globals",
    "load_module",
    EXT_PKG,
    "_override_modules",
    "_tl_modules",
    "_modules_to_import",
    "m",
    "override_module",
    "tl_module",
    "extension_module",
    "tl_package",
    "version_info",
]:
    try:
        del globals()[_n]
    except KeyError:
        pass
del globals()["_n"]

from .version import metaflow_version as _mf_version

__version__ = _mf_version



from .conda_environment import CondaEnvironment


# To placate people who don't want to see a shred of conda in UX, we symlink
# --environment=pypi to --environment=conda
class PyPIEnvironment(CondaEnvironment):
    TYPE = "pypi"



import traceback

from metaflow.exception import MetaflowException, MetaflowExceptionWrapper
from metaflow.decorators import StepDecorator
from metaflow import current

NUM_FALLBACK_RETRIES = 3


class FailureHandledByCatch(MetaflowException):
    headline = "Task execution failed but @catch handled it"

    def __init__(self, retry_count):
        msg = (
            "Task execution kept failing over %d attempts. "
            "Your code did not raise an exception. Something "
            "in the execution environment caused the failure." % retry_count
        )
        super(FailureHandledByCatch, self).__init__(msg)


class CatchDecorator(StepDecorator):
    """
    Specifies that the step will success under all circumstances.

    The decorator will create an optional artifact, specified by `var`, which
    contains the exception raised. You can use it to detect the presence
    of errors, indicating that all happy-path artifacts produced by the step
    are missing.

    Parameters
    ----------
    var : str, optional, default None
        Name of the artifact in which to store the caught exception.
        If not specified, the exception is not stored.
    print_exception : bool, default True
        Determines whether or not the exception is printed to
        stdout when caught.
    """

    name = "catch"
    defaults = {"var": None, "print_exception": True}

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # handling _foreach_var and _foreach_num_splits requires some
        # deeper thinking, so let's not support that use case for now
        self.logger = logger
        if graph[step].type == "foreach":
            raise MetaflowException(
                "@catch is defined for the step *%s* "
                "but @catch is not supported in foreach "
                "split steps." % step
            )

    def _print_exception(self, step, flow):
        self.logger(head="@catch caught an exception from %s" % flow, timestamp=False)
        for line in traceback.format_exc().splitlines():
            self.logger(">  %s" % line, timestamp=False)

    def _set_var(self, flow, val):
        var = self.attributes.get("var")
        if var:
            setattr(flow, var, val)

    def task_exception(
        self, exception, step, flow, graph, retry_count, max_user_code_retries
    ):
        # Only "catch" exceptions after all retries are exhausted
        if retry_count < max_user_code_retries:
            return False

        if self.attributes["print_exception"]:
            self._print_exception(step, flow)

        # pretend that self.next() was called as usual
        flow._transition = (graph[step].out_funcs, None)

        # If this task is a UBF control task, it will return itself as the singleton
        # list of tasks.
        if hasattr(flow, "_parallel_ubf_iter"):
            flow._control_mapper_tasks = [
                "/".join((current.run_id, current.step_name, current.task_id))
            ]
        # store the exception
        picklable = MetaflowExceptionWrapper(exception)
        flow._catch_exception = picklable
        self._set_var(flow, picklable)
        return True

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        # there was no exception, set the exception var (if any) to None
        self._set_var(flow, None)

    def step_task_retry_count(self):
        return 0, NUM_FALLBACK_RETRIES

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        # if the user code has failed max_user_code_retries times, @catch
        # runs a piece of fallback code instead. This way we can continue
        # running the flow downstream, as we have a proper entry for this task.

        def fallback_step(inputs=None):
            raise FailureHandledByCatch(retry_count)

        if retry_count > max_user_code_retries:
            return fallback_step
        else:
            return step_func



import re
import json

from metaflow import current
from metaflow.decorators import FlowDecorator
from metaflow.exception import MetaflowException
from metaflow.util import is_stringish
from metaflow.parameters import DeployTimeField, deploy_time_eval

# TODO: Support dynamic parameter mapping through a context object that exposes
#       flow name and user name similar to parameter context


class TriggerDecorator(FlowDecorator):
    """
    Specifies the event(s) that this flow depends on.

    ```
    @trigger(event='foo')
    ```
    or
    ```
    @trigger(events=['foo', 'bar'])
    ```

    Additionally, you can specify the parameter mappings
    to map event payload to Metaflow parameters for the flow.
    ```
    @trigger(event={'name':'foo', 'parameters':{'flow_param': 'event_field'}})
    ```
    or
    ```
    @trigger(events=[{'name':'foo', 'parameters':{'flow_param_1': 'event_field_1'},
                     {'name':'bar', 'parameters':{'flow_param_2': 'event_field_2'}])
    ```

    'parameters' can also be a list of strings and tuples like so:
    ```
    @trigger(event={'name':'foo', 'parameters':['common_name', ('flow_param', 'event_field')]})
    ```
    This is equivalent to:
    ```
    @trigger(event={'name':'foo', 'parameters':{'common_name': 'common_name', 'flow_param': 'event_field'}})
    ```

    Parameters
    ----------
    event : Union[str, Dict[str, Any]], optional, default None
        Event dependency for this flow.
    events : List[Union[str, Dict[str, Any]]], default []
        Events dependency for this flow.
    options : Dict[str, Any], default {}
        Backend-specific configuration for tuning eventing behavior.

    MF Add To Current
    -----------------
    trigger -> metaflow.events.Trigger
        Returns `Trigger` if the current run is triggered by an event

        @@ Returns
        -------
        Trigger
            `Trigger` if triggered by an event
    """

    name = "trigger"
    defaults = {
        "event": None,
        "events": [],
        "options": {},
    }

    def process_event_name(self, event):
        if is_stringish(event):
            return {"name": str(event)}
        elif isinstance(event, dict):
            if "name" not in event:
                raise MetaflowException(
                    "The *event* attribute for *@trigger* is missing the *name* key."
                )
            if callable(event["name"]) and not isinstance(
                event["name"], DeployTimeField
            ):
                event["name"] = DeployTimeField(
                    "event_name", str, None, event["name"], False
                )
            event["parameters"] = self.process_parameters(event.get("parameters", {}))
            return event
        elif callable(event) and not isinstance(event, DeployTimeField):
            return DeployTimeField("event", [str, dict], None, event, False)
        else:
            raise MetaflowException(
                "Incorrect format for *event* attribute in *@trigger* decorator. "
                "Supported formats are string and dictionary - \n"
                "@trigger(event='foo') or @trigger(event={'name': 'foo', "
                "'parameters': {'alpha': 'beta'}})"
            )

    def process_parameters(self, parameters):
        new_param_values = {}
        if isinstance(parameters, (list, tuple)):
            for mapping in parameters:
                if is_stringish(mapping):
                    new_param_values[mapping] = mapping
                elif callable(mapping) and not isinstance(mapping, DeployTimeField):
                    mapping = DeployTimeField(
                        "parameter_val", str, None, mapping, False
                    )
                    new_param_values[mapping] = mapping
                elif isinstance(mapping, (list, tuple)) and len(mapping) == 2:
                    if callable(mapping[0]) and not isinstance(
                        mapping[0], DeployTimeField
                    ):
                        mapping[0] = DeployTimeField(
                            "parameter_val", str, None, mapping[0], False
                        )
                    if callable(mapping[1]) and not isinstance(
                        mapping[1], DeployTimeField
                    ):
                        mapping[1] = DeployTimeField(
                            "parameter_val", str, None, mapping[1], False
                        )
                    new_param_values[mapping[0]] = mapping[1]
                else:
                    raise MetaflowException(
                        "The *parameters* attribute for event is invalid. "
                        "It should be a list/tuple of strings and lists/tuples of size 2"
                    )
        elif callable(parameters) and not isinstance(parameters, DeployTimeField):
            return DeployTimeField(
                "parameters", [list, dict, tuple], None, parameters, False
            )
        elif isinstance(parameters, dict):
            for key, value in parameters.items():
                if callable(key) and not isinstance(key, DeployTimeField):
                    key = DeployTimeField("flow_parameter", str, None, key, False)
                if callable(value) and not isinstance(value, DeployTimeField):
                    value = DeployTimeField("signal_parameter", str, None, value, False)
                new_param_values[key] = value
        return new_param_values

    def flow_init(
        self,
        flow_name,
        graph,
        environment,
        flow_datastore,
        metadata,
        logger,
        echo,
        options,
    ):
        self.triggers = []
        if sum(map(bool, (self.attributes["event"], self.attributes["events"]))) > 1:
            raise MetaflowException(
                "Specify only one of *event* or *events* "
                "attributes in *@trigger* decorator."
            )
        elif self.attributes["event"]:
            event = self.attributes["event"]
            processed_event = self.process_event_name(event)
            self.triggers.append(processed_event)
        elif self.attributes["events"]:
            # events attribute supports the following formats -
            #     1. events=[{'name': 'table.prod_db.members',
            #               'parameters': {'alpha': 'member_weight'}},
            #                {'name': 'table.prod_db.metadata',
            #               'parameters': {'beta': 'grade'}}]
            if isinstance(self.attributes["events"], list):
                # process every event in events
                for event in self.attributes["events"]:
                    processed_event = self.process_event_name(event)
                    self.triggers.append(processed_event)
            elif callable(self.attributes["events"]) and not isinstance(
                self.attributes["events"], DeployTimeField
            ):
                trig = DeployTimeField(
                    "events", list, None, self.attributes["events"], False
                )
                self.triggers.append(trig)
            else:
                raise MetaflowException(
                    "Incorrect format for *events* attribute in *@trigger* decorator. "
                    "Supported format is list - \n"
                    "@trigger(events=[{'name': 'foo', 'parameters': {'alpha': "
                    "'beta'}}, {'name': 'bar', 'parameters': "
                    "{'gamma': 'kappa'}}])"
                )

        if not self.triggers:
            raise MetaflowException("No event(s) specified in *@trigger* decorator.")

        # same event shouldn't occur more than once
        names = [
            x["name"]
            for x in self.triggers
            if not isinstance(x, DeployTimeField)
            and not isinstance(x["name"], DeployTimeField)
        ]
        if len(names) != len(set(names)):
            raise MetaflowException(
                "Duplicate event names defined in *@trigger* decorator."
            )

        self.options = self.attributes["options"]

        # TODO: Handle scenario for local testing using --trigger.

    def format_deploytime_value(self):
        new_triggers = []
        for trigger in self.triggers:
            # Case where trigger is a function that returns a list of events
            # Need to do this bc we need to iterate over list later
            if isinstance(trigger, DeployTimeField):
                evaluated_trigger = deploy_time_eval(trigger)
                if isinstance(evaluated_trigger, dict):
                    trigger = evaluated_trigger
                elif isinstance(evaluated_trigger, str):
                    trigger = {"name": evaluated_trigger}
                if isinstance(evaluated_trigger, list):
                    for trig in evaluated_trigger:
                        if is_stringish(trig):
                            new_triggers.append({"name": trig})
                        else:  # dict or another deploytimefield
                            new_triggers.append(trig)
                else:
                    new_triggers.append(trigger)
            else:
                new_triggers.append(trigger)

        self.triggers = new_triggers
        for trigger in self.triggers:
            old_trigger = trigger
            trigger_params = trigger.get("parameters", {})
            # Case where param is a function (can return list or dict)
            if isinstance(trigger_params, DeployTimeField):
                trigger_params = deploy_time_eval(trigger_params)
            # If params is a list of strings, convert to dict with same key and value
            if isinstance(trigger_params, (list, tuple)):
                new_trigger_params = {}
                for mapping in trigger_params:
                    if is_stringish(mapping) or callable(mapping):
                        new_trigger_params[mapping] = mapping
                    elif callable(mapping) and not isinstance(mapping, DeployTimeField):
                        mapping = DeployTimeField(
                            "parameter_val", str, None, mapping, False
                        )
                        new_trigger_params[mapping] = mapping
                    elif isinstance(mapping, (list, tuple)) and len(mapping) == 2:
                        if callable(mapping[0]) and not isinstance(
                            mapping[0], DeployTimeField
                        ):
                            mapping[0] = DeployTimeField(
                                "parameter_val",
                                str,
                                None,
                                mapping[1],
                                False,
                            )
                        if callable(mapping[1]) and not isinstance(
                            mapping[1], DeployTimeField
                        ):
                            mapping[1] = DeployTimeField(
                                "parameter_val",
                                str,
                                None,
                                mapping[1],
                                False,
                            )

                        new_trigger_params[mapping[0]] = mapping[1]
                    else:
                        raise MetaflowException(
                            "The *parameters* attribute for event '%s' is invalid. "
                            "It should be a list/tuple of strings and lists/tuples "
                            "of size 2" % self.attributes["event"]["name"]
                        )
                trigger_params = new_trigger_params
            trigger["parameters"] = trigger_params

            trigger_name = trigger.get("name")
            # Case where just the name is a function (always a str)
            if isinstance(trigger_name, DeployTimeField):
                trigger_name = deploy_time_eval(trigger_name)
                trigger["name"] = trigger_name

            # Third layer
            # {name:, parameters:[func, ..., ...]}
            # {name:, parameters:{func : func2}}
            for trigger in self.triggers:
                old_trigger = trigger
                trigger_params = trigger.get("parameters", {})
                new_trigger_params = {}
                for key, value in trigger_params.items():
                    if isinstance(value, DeployTimeField) and key is value:
                        evaluated_param = deploy_time_eval(value)
                        new_trigger_params[evaluated_param] = evaluated_param
                    elif isinstance(value, DeployTimeField):
                        new_trigger_params[key] = deploy_time_eval(value)
                    elif isinstance(key, DeployTimeField):
                        new_trigger_params[deploy_time_eval(key)] = value
                    else:
                        new_trigger_params[key] = value
                trigger["parameters"] = new_trigger_params
            self.triggers[self.triggers.index(old_trigger)] = trigger


class TriggerOnFinishDecorator(FlowDecorator):
    """
    Specifies the flow(s) that this flow depends on.

    ```
    @trigger_on_finish(flow='FooFlow')
    ```
    or
    ```
    @trigger_on_finish(flows=['FooFlow', 'BarFlow'])
    ```
    This decorator respects the @project decorator and triggers the flow
    when upstream runs within the same namespace complete successfully

    Additionally, you can specify project aware upstream flow dependencies
    by specifying the fully qualified project_flow_name.
    ```
    @trigger_on_finish(flow='my_project.branch.my_branch.FooFlow')
    ```
    or
    ```
    @trigger_on_finish(flows=['my_project.branch.my_branch.FooFlow', 'BarFlow'])
    ```

    You can also specify just the project or project branch (other values will be
    inferred from the current project or project branch):
    ```
    @trigger_on_finish(flow={"name": "FooFlow", "project": "my_project", "project_branch": "branch"})
    ```

    Note that `branch` is typically one of:
      - `prod`
      - `user.bob`
      - `test.my_experiment`
      - `prod.staging`

    Parameters
    ----------
    flow : Union[str, Dict[str, str]], optional, default None
        Upstream flow dependency for this flow.
    flows : List[Union[str, Dict[str, str]]], default []
        Upstream flow dependencies for this flow.
    options : Dict[str, Any], default {}
        Backend-specific configuration for tuning eventing behavior.

    MF Add To Current
    -----------------
    trigger -> metaflow.events.Trigger
        Returns `Trigger` if the current run is triggered by an event

        @@ Returns
        -------
        Trigger
            `Trigger` if triggered by an event
    """

    name = "trigger_on_finish"

    options = {
        "trigger": dict(
            multiple=True,
            default=None,
            help="Specify run pathspec for testing @trigger_on_finish locally.",
        ),
    }
    defaults = {
        "flow": None,  # flow_name or project_flow_name
        "flows": [],  # flow_names or project_flow_names
        "options": {},
        # Re-enable if you want to support TL options directly in the decorator like
        # for @project decorator
        #    **{k: v["default"] for k, v in options.items()},
    }

    def flow_init(
        self,
        flow_name,
        graph,
        environment,
        flow_datastore,
        metadata,
        logger,
        echo,
        options,
    ):
        self.triggers = []
        if sum(map(bool, (self.attributes["flow"], self.attributes["flows"]))) > 1:
            raise MetaflowException(
                "Specify only one of *flow* or *flows* "
                "attributes in *@trigger_on_finish* decorator."
            )
        elif self.attributes["flow"]:
            # flow supports the format @trigger_on_finish(flow='FooFlow')
            flow = self.attributes["flow"]
            if callable(flow) and not isinstance(
                self.attributes["flow"], DeployTimeField
            ):
                trig = DeployTimeField("fq_name", [str, dict], None, flow, False)
                self.triggers.append(trig)
            else:
                self.triggers.extend(self._parse_static_triggers([flow]))
        elif self.attributes["flows"]:
            # flows attribute supports the following formats -
            #     1. flows=['FooFlow', 'BarFlow']
            flows = self.attributes["flows"]
            if callable(flows) and not isinstance(flows, DeployTimeField):
                trig = DeployTimeField("flows", list, None, flows, False)
                self.triggers.append(trig)
            elif isinstance(flows, list):
                self.triggers.extend(self._parse_static_triggers(flows))
            else:
                raise MetaflowException(
                    "Incorrect type for *flows* attribute in *@trigger_on_finish* "
                    "decorator. Supported type is list - \n"
                    "@trigger_on_finish(flows=['FooFlow', 'BarFlow']"
                )

        if not self.triggers:
            raise MetaflowException(
                "No flow(s) specified in *@trigger_on_finish* decorator."
            )

        # Make triggers @project aware
        for trigger in self.triggers:
            if isinstance(trigger, DeployTimeField):
                continue
            self._parse_fq_name(trigger)

        self.options = self.attributes["options"]

        # Handle scenario for local testing using --trigger.

        # Re-enable this code if you want to support passing trigger directly in the
        # decorator in a way similar to how production and branch are passed in the
        # project decorator.

        # # This is overkill since default is None for all options but adding this code
        # # to make it safe if other non None-default options are added in the future.
        # for op in options:
        #     if (
        #         op in self._user_defined_attributes
        #         and options[op] != self.defaults[op]
        #         and self.attributes[op] != options[op]
        #     ):
        #         # Exception if:
        #         #  - the user provides a value in the attributes field
        #         #  - AND the user provided a value in the command line (non default)
        #         #  - AND the values are different
        #         # Note that this won't raise an error if the user provided the default
        #         # value in the command line and provided one in attribute but although
        #         # slightly inconsistent, it is not incorrect.
        #         raise MetaflowException(
        #             "You cannot pass %s as both a command-line argument and an attribute "
        #             "of the @trigger_on_finish decorator." % op
        #         )

        # if "trigger" in self._user_defined_attributes:
        #    trigger_option = self.attributes["trigger"]
        # else:
        trigger_option = options["trigger"]

        self._option_values = options
        if trigger_option:
            from metaflow import Run
            from metaflow.events import Trigger

            run_objs = []
            for run_pathspec in trigger_option:
                if len(run_pathspec.split("/")) != 2:
                    raise MetaflowException(
                        "Incorrect format for run pathspec for *--trigger*. "
                        "Supported format is flow_name/run_id."
                    )
                run_obj = Run(run_pathspec, _namespace_check=False)
                if not run_obj.successful:
                    raise MetaflowException(
                        "*--trigger* does not support runs that are not successful yet."
                    )
                run_objs.append(run_obj)
            current._update_env({"trigger": Trigger.from_runs(run_objs)})

    @staticmethod
    def _parse_static_triggers(flows):
        results = []
        for flow in flows:
            if is_stringish(flow):
                results.append(
                    {
                        "fq_name": flow,
                    }
                )
            elif isinstance(flow, dict):
                if "name" not in flow:
                    if len(flows) > 1:
                        raise MetaflowException(
                            "One or more flows in the *flows* attribute for "
                            "*@trigger_on_finish* is missing the "
                            "*name* key."
                        )
                    raise MetaflowException(
                        "The *flow* attribute for *@trigger_on_finish* is missing the "
                        "*name* key."
                    )
                flow_name = flow["name"]

                if not is_stringish(flow_name) or "." in flow_name:
                    raise MetaflowException(
                        f"The *name* attribute of the *flow* {flow_name} is not a valid string"
                    )
                result = {"fq_name": flow_name}
                if "project" in flow:
                    if is_stringish(flow["project"]):
                        result["project"] = flow["project"]
                    else:
                        raise MetaflowException(
                            f"The *project* attribute of the *flow* {flow_name} is not a string"
                        )
                if "project_branch" in flow:
                    if is_stringish(flow["project_branch"]):
                        result["branch"] = flow["project_branch"]
                    else:
                        raise MetaflowException(
                            f"The *project_branch* attribute of the *flow* {flow_name} is not a string"
                        )
                results.append(result)
            else:
                if len(flows) > 1:
                    raise MetaflowException(
                        "One or more flows in the *flows* attribute for "
                        "*@trigger_on_finish* decorator have an incorrect type. "
                        "Supported type is string or Dict[str, str]- \n"
                        "@trigger_on_finish(flows=['FooFlow', 'BarFlow']"
                    )
                raise MetaflowException(
                    "Incorrect type for *flow* attribute in *@trigger_on_finish* "
                    " decorator. Supported type is string or Dict[str, str] - \n"
                    "@trigger_on_finish(flow='FooFlow') or "
                    "@trigger_on_finish(flow={'name':'FooFlow', 'project_branch': 'branch'})"
                )
        return results

    def _parse_fq_name(self, trigger):
        if trigger["fq_name"].count(".") == 0:
            # fully qualified name is just the flow name
            trigger["flow"] = trigger["fq_name"]
        elif trigger["fq_name"].count(".") >= 2:
            # fully qualified name is of the format - project.branch.flow_name
            trigger["project"], tail = trigger["fq_name"].split(".", maxsplit=1)
            trigger["branch"], trigger["flow"] = tail.rsplit(".", maxsplit=1)
        else:
            raise MetaflowException(
                "Incorrect format for *flow* in *@trigger_on_finish* "
                "decorator. Specify either just the *flow_name* or a fully "
                "qualified name like *project_name.branch_name.flow_name*."
            )
        if not re.match(r"^[A-Za-z0-9_]+$", trigger["flow"]):
            raise MetaflowException(
                "Invalid flow name *%s* in *@trigger_on_finish* "
                "decorator. Only alphanumeric characters and "
                "underscores(_) are allowed." % trigger["flow"]
            )

    def format_deploytime_value(self):
        if len(self.triggers) == 1 and isinstance(self.triggers[0], DeployTimeField):
            deploy_value = deploy_time_eval(self.triggers[0])
            if isinstance(deploy_value, list):
                self.triggers = deploy_value
            else:
                self.triggers = [deploy_value]
            triggers = self._parse_static_triggers(self.triggers)
            for trigger in triggers:
                self._parse_fq_name(trigger)
            self.triggers = triggers

    def get_top_level_options(self):
        return list(self._option_values.items())



from metaflow.sidecar import Message, MessageTypes, Sidecar


class NullEventLogger(object):
    TYPE = "nullSidecarLogger"

    def __init__(self, *args, **kwargs):
        # Currently passed flow and env in kwargs
        self._sidecar = Sidecar(self.TYPE)

    def start(self):
        return self._sidecar.start()

    def terminate(self):
        return self._sidecar.terminate()

    def send(self, msg):
        # Arbitrary message sending. Useful if you want to override some different
        # types of messages.
        self._sidecar.send(msg)

    def log(self, payload):
        if self._sidecar.is_active:
            msg = Message(MessageTypes.BEST_EFFORT, payload)
            self._sidecar.send(msg)

    @classmethod
    def get_worker(cls):
        return None



import json
import sys
import base64
import datetime
from collections import namedtuple

TypeResolvedObject = namedtuple("TypeResolvedObject", ["data", "is_image", "is_table"])


TIME_FORMAT = "%Y-%m-%d %I:%M:%S %p"
MAX_ARTIFACT_SIZE = 1  # in 1 MB


def _get_object_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([_get_object_size(v, seen) for v in obj.values()])
        size += sum([_get_object_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([_get_object_size(i, seen) for i in obj])
        if hasattr(obj, "__dict__"):
            size += _get_object_size(obj.__dict__.values(), seen)
    elif hasattr(obj, "__dict__"):
        size += _get_object_size(obj.__dict__, seen)
    return size


def _full_classname(obj):
    cls = type(obj)
    module = cls.__module__
    name = cls.__qualname__
    if module is not None and module != "__builtin__" and module != "builtins":
        name = module + "." + name
    return name


class TaskToDict:
    def __init__(self, only_repr=False, runtime=False):
        # this dictionary holds all the supported functions
        import reprlib
        import pprint

        self._pretty_print = pprint

        r = reprlib.Repr()
        r.maxarray = 100
        r.maxstring = 100
        r.maxother = 100
        r.maxtuple = 100
        r.maxlist = 100
        r.maxlevel = 3
        self._repr = r
        self._runtime = runtime
        self._only_repr = only_repr
        self._supported_types = {
            "tuple": self._parse_tuple,
            "NoneType": self._parse_nonetype,
            "set": self._parse_set,
            "frozenset": self._parse_frozenset,
            "bytearray": self._parse_bytearray,
            "str": self._parse_str,
            "datetime.datetime": self._parse_datetime_datetime,
            "bool": self._parse_bool,
            "decimal.Decimal": self._parse_decimal_decimal,
            "type": self._parse_type,
            "range": self._parse_range,
            "pandas.core.frame.DataFrame": self._parse_pandas_dataframe,
            "numpy.ndarray": self._parse_numpy_ndarray,
            "dict": self._parse_dict,
            "float": self._parse_float,
            "complex": self._parse_complex,
            "int": self._parse_int,
            "Exception": self._parse_exception,
            "list": self._parse_list,
            "bytes": self._parse_bytes,
        }

    def __call__(self, task, graph=None):
        # task_props = ['stderr','stdout','created_at','finished_at','pathspec']
        # todo : dictionary Pretty printing.
        task_dict = dict(
            stderr=task.stderr,
            stdout=task.stdout,
            created_at=task.created_at.strftime(TIME_FORMAT),
            finished_at=None,
            pathspec=task.pathspec,
            graph=graph,
            data={},
        )
        if not self._runtime:
            if task.finished_at is not None:
                task_dict.update(
                    dict(finished_at=task.finished_at.strftime(TIME_FORMAT))
                )
        task_dict["data"], type_infered_objects = self._create_task_data_dict(task)
        task_dict.update(type_infered_objects)
        return task_dict

    def _create_task_data_dict(self, task):

        task_data_dict = {}
        type_inferred_objects = {"images": {}, "tables": {}}
        for data in task:
            try:
                data_object = data.data
                task_data_dict[data.id] = self._convert_to_native_type(data_object)
                task_data_dict[data.id]["name"] = data.id
            except ModuleNotFoundError as e:
                data_object = "<unable to unpickle>"
                # this means pickle couldn't find the module.
                task_data_dict[data.id] = dict(
                    type=e.name,
                    data=data_object,
                    large_object=False,
                    supported_type=False,
                    only_repr=self._only_repr,
                    name=data.id,
                )

            # Resolve special types.
            type_resolved_obj = self._extract_type_infered_object(data_object)
            if type_resolved_obj is not None:
                if type_resolved_obj.is_image:
                    type_inferred_objects["images"][data.id] = type_resolved_obj.data
                elif type_resolved_obj.is_table:
                    type_inferred_objects["tables"][data.id] = type_resolved_obj.data

        return task_data_dict, type_inferred_objects

    def object_type(self, object):
        return self._get_object_type(object)

    def parse_image(self, data_object):
        obj_type_name = self._get_object_type(data_object)
        if obj_type_name == "bytes":
            # Works for python 3.1+
            # Python 3.13 removes the standard ``imghdr`` module. Metaflow
            # vendors a copy so we can keep using ``what`` to detect image
            # formats irrespective of the Python version.
            import warnings

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=DeprecationWarning, module="imghdr"
                )
                from metaflow._vendor import imghdr

            resp = imghdr.what(None, h=data_object)
            # Only accept types supported on the web
            # https://developer.mozilla.org/en-US/docs/Web/Media/Formats/Image_types
            if resp is not None and resp in ["gif", "png", "jpeg", "webp"]:
                return self._parse_image(data_object, resp)
        return None

    def _extract_type_infered_object(self, data_object):
        # check images
        obj_type_name = self._get_object_type(data_object)
        if obj_type_name == "bytes":
            # Works for python 3.1+
            from metaflow._vendor import imghdr

            resp = imghdr.what(None, h=data_object)
            # Only accept types supported on the web
            # https://developer.mozilla.org/en-US/docs/Web/Media/Formats/Image_types
            if resp is not None and resp in ["gif", "png", "jpeg", "webp"]:
                return TypeResolvedObject(
                    self._parse_image(data_object, resp), True, False
                )
        elif obj_type_name == "pandas.core.frame.DataFrame":
            return TypeResolvedObject(
                self._parse_pandas_dataframe(data_object), False, True
            )
        return None

    def _parse_image(self, dataobject, img_type):
        return "data:image/%s;base64, %s" % (
            img_type.lower(),
            self._parse_bytes(dataobject),
        )

    @staticmethod
    def _get_object_type(obj_val):
        """returns string or None"""
        try:
            return _full_classname(obj_val)
        except AttributeError as e:
            pass

        return None

    def infer_object(self, artifact_object):
        return self._convert_to_native_type(artifact_object)

    def _convert_to_native_type(
        self,
        artifact_object,
    ):
        # For he current iteration return a dictionary.
        #
        data_dict = dict(
            type=None,
            data=None,
            large_object=False,
            supported_type=False,
            only_repr=self._only_repr,
        )
        (
            data_dict["data"],
            data_dict["type"],
            data_dict["supported_type"],
            data_dict["large_object"],
        ) = self._to_native_type(artifact_object)
        # when obj_type_dict name is none means type was not resolvable.
        return data_dict

    def _to_native_type(self, data_object):
        # returns data_obj, obj_type_name, supported_type, large_object
        rep = self._get_repr()
        supported_type = False
        large_object = False
        obj_type_name = self._get_object_type(data_object)
        if obj_type_name is None:
            return rep.repr(data_object), obj_type_name, supported_type, large_object
        elif self._only_repr:
            return (
                self._pretty_print_obj(data_object),
                obj_type_name,
                supported_type,
                large_object,
            )
        if obj_type_name in self._supported_types:
            supported_type = True
            type_parsing_func = self._supported_types[obj_type_name]
            data_obj = type_parsing_func(data_object)
            if _get_object_size(data_obj) * 1e-6 > MAX_ARTIFACT_SIZE:
                data_obj = rep.repr(data_obj)
                large_object = True
        else:
            # If object is not in supported types get its REPR
            data_obj = rep.repr(data_object)

        return data_obj, obj_type_name, supported_type, large_object

    def _pretty_print_obj(self, data_object):
        data = self._repr.repr(data_object)
        if "..." in data:
            return data
        else:
            pretty_print_op = self._pretty_print.pformat(
                data_object, indent=2, width=50, compact=True
            )
            if pretty_print_op is None:
                return data
            return pretty_print_op

    def _get_repr(self):
        return self._repr

    def _parse_tuple(self, data_object):
        return self._parse_list([obj for obj in data_object])

    def _parse_nonetype(self, data_object):
        return data_object

    def _parse_set(self, data_object):
        return self._parse_frozenset(data_object)

    def _parse_frozenset(self, data_object):
        ret_vals = []
        for obj in list(data_object):
            (
                data_obj,
                obj_type_name,
                supported_type,
                large_object,
            ) = self._to_native_type(obj)
            ret_vals.append(data_obj)
        return ret_vals

    def _parse_bytearray(self, data_object):
        try:
            return data_object.decode("utf-8")
        except UnicodeDecodeError as e:
            return self._get_repr().repr(data_object)

    def _parse_str(self, data_object):
        return data_object

    def _parse_datetime_datetime(self, data_object):
        return data_object.strftime(TIME_FORMAT)

    def _parse_bool(self, data_object):
        return data_object

    def _parse_decimal_decimal(self, data_object):
        return float(data_object)

    def _parse_type(self, data_object):
        return data_object.__name__

    def _parse_range(self, data_object):
        return self._get_repr().repr(data_object)

    @staticmethod
    def _parse_pandas_column(column_object):
        # There are two types of parsing we do here.
        # 1. We explicitly parse the types we know how to parse
        # 2. We try to partially match a type name to the column's type.
        #   - We do this because `datetime64` can match `datetime64[ns]` and `datetime64[ns, UTC]`
        #   - We do this because period can match `period[D]` and `period[2D]` etc.
        #   - There are just too many types to explicitly parse so we go by this heuristic
        # We have a default parser called `truncate_long_objects` which type casts any column to string
        # and truncates it to 30 characters.
        # If there is any form of TypeError or ValueError we set the column value to "Unsupported Type"
        # We also set columns which are have null values to "null" strings
        time_format = "%Y-%m-%dT%H:%M:%S%Z"
        truncate_long_objects = lambda x: (
            x.astype("string").str.slice(0, 30) + "..."
            if len(x) > 0 and x.astype("string").str.len().max() > 30
            else x.astype("string")
        )
        type_parser = {
            "int64": lambda x: x,
            "float64": lambda x: x,
            "bool": lambda x: x,
            "object": lambda x: truncate_long_objects(x.fillna("null")),
            "category": truncate_long_objects,
        }

        partial_type_name_match_parsers = {
            "complex": {
                "complex": lambda x: x.astype("string"),
            },
            "datetime": {
                "datetime64": lambda x: x.dt.strftime(time_format),
                "timedelta": lambda x: x.dt.total_seconds(),
            },
            "interval": {
                "interval": lambda x: x.astype("string"),
            },
            "period": {
                "period": lambda x: x.astype("string"),
            },
        }

        def _match_partial_type():
            col_type = column_object.dtype
            for _, type_parsers in partial_type_name_match_parsers.items():
                for type_name, parser in type_parsers.items():
                    if type_name in str(col_type):
                        return parser(column_object)
            return None

        try:
            col_type = str(column_object.dtype)
            if col_type in type_parser:
                return type_parser[col_type](column_object.fillna("null"))
            else:
                parsed_col = _match_partial_type()
                if parsed_col is not None:
                    return parsed_col.fillna("null")
            return truncate_long_objects(column_object.fillna("null"))
        except ValueError as e:
            return "Unsupported type: {0}".format(col_type)
        except TypeError as e:
            return "Unsupported type: {0}".format(col_type)

    def _parse_pandas_dataframe(self, data_object, truncate=True):
        headers = list(data_object.columns)
        data = data_object
        if truncate:
            data = data_object.head()
        index_column = data.index

        # We explicitly cast the `index_column` object to an `Index` or `MultiIndex` having JSON-castable values.
        if index_column.__class__.__name__ == "MultiIndex":
            from pandas import MultiIndex

            cols = [
                self._parse_pandas_column(
                    index_column.get_level_values(name).to_series()
                )
                for name in index_column.names
            ]
            index_column = MultiIndex.from_arrays(cols, names=index_column.names)
        else:
            from pandas import Index

            index_column = Index(self._parse_pandas_column(index_column.to_series()))

        for col in data.columns:
            data[col] = self._parse_pandas_column(data[col])

        data_vals = data.values.tolist()
        for row, idx in zip(data_vals, index_column.values.tolist()):
            row.insert(0, idx)
        return dict(
            full_size=(
                # full_size is a tuple of (num_rows,num_columns)
                len(data_object),
                len(headers),
            ),
            headers=[""] + headers,
            data=data_vals,
            truncated=truncate,
        )

    def _parse_numpy_ndarray(self, data_object):
        return data_object.tolist()

    def _parse_dict(self, data_object):
        data_dict = {}
        for d in data_object:
            (
                data_obj,
                obj_type_name,
                supported_type,
                large_object,
            ) = self._to_native_type(data_object[d])
            data_dict[d] = data_obj
        return data_dict

    def _parse_float(self, data_object):
        return data_object

    def _parse_complex(self, data_object):
        return str(data_object)

    def _parse_int(self, data_object):
        return data_object

    def _parse_exception(self, data_object):
        repr = self._get_repr()
        return repr.repr(data_object)

    def _parse_list(self, data_object):
        data_list = []
        for obj in data_object:
            data_obj, _, _, _ = self._to_native_type(obj)
            data_list.append(data_obj)
        return data_list

    def _parse_bytes(self, data_object):
        # encode bytes to base64 as they maybe images.
        return base64.encodebytes(data_object).decode("utf8")



import re
from typing import Dict, List, Optional
from metaflow.exception import CommandException, MetaflowException
from metaflow.util import get_username, get_latest_run_id


# avoid circular import by having the exception class contained here
class KubernetesException(MetaflowException):
    headline = "Kubernetes error"


def parse_cli_options(flow_name, run_id, user, my_runs, echo):
    if user and my_runs:
        raise CommandException("--user and --my-runs are mutually exclusive.")

    if run_id and my_runs:
        raise CommandException("--run_id and --my-runs are mutually exclusive.")

    if my_runs:
        user = get_username()

    latest_run = True

    if user and not run_id:
        latest_run = False

    if not run_id and latest_run:
        run_id = get_latest_run_id(echo, flow_name)
        if run_id is None:
            raise CommandException("A previous run id was not found. Specify --run-id.")

    return flow_name, run_id, user


def qos_requests_and_limits(qos: str, cpu: int, memory: int, storage: int):
    "return resource requests and limits for the kubernetes pod based on the given QoS Class"
    # case insensitive matching for QoS class
    qos = qos.lower()
    # Determine the requests and limits to define chosen QoS class
    qos_limits = {}
    qos_requests = {}
    if qos == "guaranteed":
        # Guaranteed - has both cpu/memory limits. requests not required, as these will be inferred.
        qos_limits = {
            "cpu": str(cpu),
            "memory": "%sM" % str(memory),
            "ephemeral-storage": "%sM" % str(storage),
        }
        # NOTE: Even though Kubernetes will produce matching requests for the specified limits, this happens late in the lifecycle.
        # We specify them explicitly here to make some K8S tooling happy, in case they rely on .resources.requests being present at time of submitting the job.
        qos_requests = qos_limits
    else:
        # Burstable - not Guaranteed, and has a memory/cpu limit or request
        qos_requests = {
            "cpu": str(cpu),
            "memory": "%sM" % str(memory),
            "ephemeral-storage": "%sM" % str(storage),
        }
    # TODO: Add support for BestEffort once there is a use case for it.
    # BestEffort - no limit or requests for cpu/memory
    return qos_requests, qos_limits


def validate_kube_labels(
    labels: Optional[Dict[str, Optional[str]]],
) -> bool:
    """Validate label values.

    This validates the kubernetes label values.  It does not validate the keys.
    Ideally, keys should be static and also the validation rules for keys are
    more complex than those for values.  For full validation rules, see:

    https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set
    """

    def validate_label(s: Optional[str]):
        regex_match = r"^(([A-Za-z0-9][-A-Za-z0-9_.]{0,61})?[A-Za-z0-9])?$"
        if not s:
            # allow empty label
            return True
        if not re.search(regex_match, s):
            raise KubernetesException(
                'Invalid value: "%s"\n'
                "A valid label must be an empty string or one that\n"
                "  - Consist of alphanumeric, '-', '_' or '.' characters\n"
                "  - Begins and ends with an alphanumeric character\n"
                "  - Is at most 63 characters" % s
            )
        return True

    return all([validate_label(v) for v in labels.values()]) if labels else True


def parse_kube_keyvalue_list(items: List[str], requires_both: bool = True):
    try:
        ret = {}
        for item_str in items:
            item = item_str.split("=", 1)
            if requires_both:
                item[1]  # raise IndexError
            if str(item[0]) in ret:
                raise KubernetesException("Duplicate key found: %s" % str(item[0]))
            ret[str(item[0])] = str(item[1]) if len(item) > 1 else None
        return ret
    except KubernetesException as e:
        raise e
    except (AttributeError, IndexError):
        raise KubernetesException("Unable to parse kubernetes list: %s" % items)
